# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2004-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
:mod:`silx.gui.plot.actions.io` provides a set of QAction relative of inputs
and outputs for a :class:`.PlotWidget`.

The following QAction are available:

- :class:`CopyAction`
- :class:`PrintAction`
- :class:`SaveAction`
"""

from __future__ import division

__authors__ = ["V.A. Sole", "T. Vincent", "P. Knobel"]
__license__ = "MIT"
__date__ = "25/09/2020"

from . import PlotAction
from silx.io.utils import save1D, savespec, NEXUS_HDF5_EXT
from silx.io.nxdata import save_NXdata
import logging
import sys
import os.path
from collections import OrderedDict
import traceback
import numpy
from silx.utils.deprecation import deprecated
from silx.gui import qt, printer
from silx.gui.dialog.GroupDialog import GroupDialog
from silx.third_party.EdfFile import EdfFile
from silx.third_party.TiffIO import TiffIO
from ...utils.image import convertArrayToQImage
if sys.version_info[0] == 3:
    from io import BytesIO
else:
    import cStringIO as _StringIO
    BytesIO = _StringIO.StringIO

_logger = logging.getLogger(__name__)

_NEXUS_HDF5_EXT_STR = ' '.join(['*' + ext for ext in NEXUS_HDF5_EXT])


def selectOutputGroup(h5filename):
    """Open a dialog to prompt the user to select a group in
    which to output data.

    :param str h5filename: name of an existing HDF5 file
    :rtype: str
    :return: Name of output group, or None if the dialog was cancelled
    """
    dialog = GroupDialog()
    dialog.addFile(h5filename)
    dialog.setWindowTitle("Select an output group")
    if not dialog.exec_():
        return None
    return dialog.getSelectedDataUrl().data_path()


class SaveAction(PlotAction):
    """QAction for saving Plot content.

    It opens a Save as... dialog.

    :param plot: :class:`.PlotWidget` instance on which to operate.
    :param parent: See :class:`QAction`.
    """

    SNAPSHOT_FILTER_SVG = 'Plot Snapshot as SVG (*.svg)'
    SNAPSHOT_FILTER_PNG = 'Plot Snapshot as PNG (*.png)'

    DEFAULT_ALL_FILTERS = (SNAPSHOT_FILTER_PNG, SNAPSHOT_FILTER_SVG)

    # Dict of curve filters with CSV-like format
    # Using ordered dict to guarantee filters order
    # Note: '%.18e' is numpy.savetxt default format
    CURVE_FILTERS_TXT = OrderedDict((
        ('Curve as Raw ASCII (*.txt)',
         {'fmt': '%.18e', 'delimiter': ' ', 'header': False}),
        ('Curve as ";"-separated CSV (*.csv)',
         {'fmt': '%.18e', 'delimiter': ';', 'header': True}),
        ('Curve as ","-separated CSV (*.csv)',
         {'fmt': '%.18e', 'delimiter': ',', 'header': True}),
        ('Curve as tab-separated CSV (*.csv)',
         {'fmt': '%.18e', 'delimiter': '\t', 'header': True}),
        ('Curve as OMNIC CSV (*.csv)',
         {'fmt': '%.7E', 'delimiter': ',', 'header': False}),
        ('Curve as SpecFile (*.dat)',
         {'fmt': '%.10g', 'delimiter': '', 'header': False})
    ))

    CURVE_FILTER_NPY = 'Curve as NumPy binary file (*.npy)'

    CURVE_FILTER_NXDATA = 'Curve as NXdata (%s)' % _NEXUS_HDF5_EXT_STR

    DEFAULT_CURVE_FILTERS = list(CURVE_FILTERS_TXT.keys()) + [
        CURVE_FILTER_NPY, CURVE_FILTER_NXDATA]

    DEFAULT_ALL_CURVES_FILTERS = ("All curves as SpecFile (*.dat)",)

    IMAGE_FILTER_EDF = 'Image data as EDF (*.edf)'
    IMAGE_FILTER_TIFF = 'Image data as TIFF (*.tif)'
    IMAGE_FILTER_NUMPY = 'Image data as NumPy binary file (*.npy)'
    IMAGE_FILTER_ASCII = 'Image data as ASCII (*.dat)'
    IMAGE_FILTER_CSV_COMMA = 'Image data as ,-separated CSV (*.csv)'
    IMAGE_FILTER_CSV_SEMICOLON = 'Image data as ;-separated CSV (*.csv)'
    IMAGE_FILTER_CSV_TAB = 'Image data as tab-separated CSV (*.csv)'
    IMAGE_FILTER_RGB_PNG = 'Image as PNG (*.png)'
    IMAGE_FILTER_NXDATA = 'Image as NXdata (%s)' % _NEXUS_HDF5_EXT_STR

    DEFAULT_IMAGE_FILTERS = (IMAGE_FILTER_EDF,
                             IMAGE_FILTER_TIFF,
                             IMAGE_FILTER_NUMPY,
                             IMAGE_FILTER_ASCII,
                             IMAGE_FILTER_CSV_COMMA,
                             IMAGE_FILTER_CSV_SEMICOLON,
                             IMAGE_FILTER_CSV_TAB,
                             IMAGE_FILTER_RGB_PNG,
                             IMAGE_FILTER_NXDATA)

    SCATTER_FILTER_NXDATA = 'Scatter as NXdata (%s)' % _NEXUS_HDF5_EXT_STR
    DEFAULT_SCATTER_FILTERS = (SCATTER_FILTER_NXDATA,)

    # filters for which we don't want an "overwrite existing file" warning
    DEFAULT_APPEND_FILTERS = (CURVE_FILTER_NXDATA, IMAGE_FILTER_NXDATA,
                              SCATTER_FILTER_NXDATA)

    def __init__(self, plot, parent=None):
        self._filters = {
            'all': OrderedDict(),
            'curve': OrderedDict(),
            'curves': OrderedDict(),
            'image': OrderedDict(),
            'scatter': OrderedDict()}

        self._appendFilters = list(self.DEFAULT_APPEND_FILTERS)

        # Initialize filters
        for nameFilter in self.DEFAULT_ALL_FILTERS:
            self.setFileFilter(
                dataKind='all', nameFilter=nameFilter, func=self._saveSnapshot)

        for nameFilter in self.DEFAULT_CURVE_FILTERS:
            self.setFileFilter(
                dataKind='curve', nameFilter=nameFilter, func=self._saveCurve)

        for nameFilter in self.DEFAULT_ALL_CURVES_FILTERS:
            self.setFileFilter(
                dataKind='curves', nameFilter=nameFilter, func=self._saveCurves)

        for nameFilter in self.DEFAULT_IMAGE_FILTERS:
            self.setFileFilter(
                dataKind='image', nameFilter=nameFilter, func=self._saveImage)

        for nameFilter in self.DEFAULT_SCATTER_FILTERS:
            self.setFileFilter(
                dataKind='scatter', nameFilter=nameFilter, func=self._saveScatter)

        super(SaveAction, self).__init__(
            plot, icon='document-save', text='Save as...',
            tooltip='Save curve/image/plot snapshot dialog',
            triggered=self._actionTriggered,
            checkable=False, parent=parent)
        self.setShortcut(qt.QKeySequence.Save)
        self.setShortcutContext(qt.Qt.WidgetShortcut)

    @staticmethod
    def _errorMessage(informativeText='', parent=None):
        """Display an error message."""
        # TODO issue with QMessageBox size fixed and too small
        msg = qt.QMessageBox(parent)
        msg.setIcon(qt.QMessageBox.Critical)
        msg.setInformativeText(informativeText + ' ' + str(sys.exc_info()[1]))
        msg.setDetailedText(traceback.format_exc())
        msg.exec_()

    def _saveSnapshot(self, plot, filename, nameFilter):
        """Save a snapshot of the :class:`PlotWindow` widget.

        :param str filename: The name of the file to write
        :param str nameFilter: The selected name filter
        :return: False if format is not supported or save failed,
                 True otherwise.
        """
        if nameFilter == self.SNAPSHOT_FILTER_PNG:
            fileFormat = 'png'
        elif nameFilter == self.SNAPSHOT_FILTER_SVG:
            fileFormat = 'svg'
        else:  # Format not supported
            _logger.error(
                'Saving plot snapshot failed: format not supported')
            return False

        plot.saveGraph(filename, fileFormat=fileFormat)
        return True

    def _getAxesLabels(self, item):
        # If curve has no associated label, get the default from the plot
        xlabel = item.getXLabel() or self.plot.getXAxis().getLabel()
        ylabel = item.getYLabel() or self.plot.getYAxis().getLabel()
        return xlabel, ylabel

    def _get1dData(self, item):
        "provide xdata, [ydata], xlabel, [ylabel] and manages error bars"
        xlabel, ylabel = self._getAxesLabels(item)
        x_data = item.getXData(copy=False)
        y_data = item.getYData(copy=False)
        x_err = item.getXErrorData(copy=False)
        y_err = item.getYErrorData(copy=False)
        labels = [ylabel]
        data = [y_data]

        if x_err is not None:
            if numpy.isscalar(x_err):
                data.append(numpy.zeros_like(y_data) + x_err)
                labels.append(xlabel + "_errors")
            elif x_err.ndim == 1:
                data.append(x_err)
                labels.append(xlabel + "_errors")
            elif x_err.ndim == 2:
                data.append(x_err[0])
                labels.append(xlabel + "_errors_below")
                data.append(x_err[1])
                labels.append(xlabel + "_errors_above")

        if y_err is not None:
            if numpy.isscalar(y_err):
                data.append(numpy.zeros_like(y_data) + y_err)
                labels.append(ylabel + "_errors")
            elif y_err.ndim == 1:
                data.append(y_err)
                labels.append(ylabel + "_errors")
            elif y_err.ndim == 2:
                data.append(y_err[0])
                labels.append(ylabel + "_errors_below")
                data.append(y_err[1])
                labels.append(ylabel + "_errors_above")
        return x_data, data, xlabel, labels

    @staticmethod
    def _selectWriteableOutputGroup(filename, parent):
        if os.path.exists(filename) and os.path.isfile(filename) \
                and os.access(filename, os.W_OK):
            entryPath = selectOutputGroup(filename)
            if entryPath is None:
                _logger.info("Save operation cancelled")
                return None
            return entryPath
        elif not os.path.exists(filename):
            # create new entry in new file
            return "/entry"
        else:
            SaveAction._errorMessage('Save failed (file access issue)\n', parent=parent)
            return None

    def _saveCurveAsNXdata(self, curve, filename):
        entryPath = self._selectWriteableOutputGroup(filename, parent=self.plot)
        if entryPath is None:
            return False

        xlabel, ylabel = self._getAxesLabels(curve)

        return save_NXdata(
            filename,
            nxentry_name=entryPath,
            signal=curve.getYData(copy=False),
            axes=[curve.getXData(copy=False)],
            signal_name="y",
            axes_names=["x"],
            signal_long_name=ylabel,
            axes_long_names=[xlabel],
            signal_errors=curve.getYErrorData(copy=False),
            axes_errors=[curve.getXErrorData(copy=True)],
            title=self.plot.getGraphTitle())

    def _saveCurve(self, plot, filename, nameFilter):
        """Save a curve from the plot.

        :param str filename: The name of the file to write
        :param str nameFilter: The selected name filter
        :return: False if format is not supported or save failed,
                 True otherwise.
        """
        if nameFilter not in self.DEFAULT_CURVE_FILTERS:
            return False

        # Check if a curve is to be saved
        curve = plot.getActiveCurve()
        # before calling _saveCurve, if there is no selected curve, we
        # make sure there is only one curve on the graph
        if curve is None:
            curves = plot.getAllCurves()
            if not curves:
                self._errorMessage("No curve to be saved", parent=self.plot)
                return False
            curve = curves[0]

        if nameFilter in self.CURVE_FILTERS_TXT:
            filter_ = self.CURVE_FILTERS_TXT[nameFilter]
            fmt = filter_['fmt']
            csvdelim = filter_['delimiter']
            autoheader = filter_['header']
        else:
            # .npy or nxdata
            fmt, csvdelim, autoheader = ("", "", False)

        if nameFilter == self.CURVE_FILTER_NXDATA:
            return self._saveCurveAsNXdata(curve, filename)

        xdata, data, xlabel, labels = self._get1dData(curve)

        try:
            save1D(filename,
                   xdata, data,
                   xlabel, labels,
                   fmt=fmt, csvdelim=csvdelim,
                   autoheader=autoheader)
        except IOError:
            self._errorMessage('Save failed\n', parent=self.plot)
            return False

        return True

    def _saveCurves(self, plot, filename, nameFilter):
        """Save all curves from the plot.

        :param str filename: The name of the file to write
        :param str nameFilter: The selected name filter
        :return: False if format is not supported or save failed,
                 True otherwise.
        """
        if nameFilter not in self.DEFAULT_ALL_CURVES_FILTERS:
            return False

        curves = plot.getAllCurves()
        if not curves:
            self._errorMessage("No curves to be saved", parent=self.plot)
            return False

        curve = curves[0]
        scanno = 1
        try:
            xdata, data, xlabel, labels = self._get1dData(curve)

            specfile = savespec(filename,
                                xdata, data,
                                xlabel, labels,
                                fmt="%.7g", scan_number=1, mode="w",
                                write_file_header=True,
                                close_file=False)
        except IOError:
            self._errorMessage('Save failed\n', parent=self.plot)
            return False

        for curve in curves[1:]:
            try:
                scanno += 1
                xdata, data, xlabel, labels = self._get1dData(curve)
                specfile = savespec(specfile,
                                    xdata, data,
                                    xlabel, labels,
                                    fmt="%.7g", scan_number=scanno,
                                    write_file_header=False,
                                    close_file=False)
            except IOError:
                self._errorMessage('Save failed\n', parent=self.plot)
                return False
        specfile.close()

        return True

    def _saveImage(self, plot, filename, nameFilter):
        """Save an image from the plot.

        :param str filename: The name of the file to write
        :param str nameFilter: The selected name filter
        :return: False if format is not supported or save failed,
                 True otherwise.
        """
        if nameFilter not in self.DEFAULT_IMAGE_FILTERS:
            return False

        image = plot.getActiveImage()
        if image is None:
            qt.QMessageBox.warning(
                plot, "No Data", "No image to be saved")
            return False

        data = image.getData(copy=False)

        # TODO Use silx.io for writing files
        if nameFilter == self.IMAGE_FILTER_EDF:
            edfFile = EdfFile(filename, access="w+")
            edfFile.WriteImage({}, data, Append=0)
            return True

        elif nameFilter == self.IMAGE_FILTER_TIFF:
            tiffFile = TiffIO(filename, mode='w')
            tiffFile.writeImage(data, software='silx')
            return True

        elif nameFilter == self.IMAGE_FILTER_NUMPY:
            try:
                numpy.save(filename, data)
            except IOError:
                self._errorMessage('Save failed\n', parent=self.plot)
                return False
            return True

        elif nameFilter == self.IMAGE_FILTER_NXDATA:
            entryPath = self._selectWriteableOutputGroup(filename, parent=self.plot)
            if entryPath is None:
                return False
            xorigin, yorigin = image.getOrigin()
            xscale, yscale = image.getScale()
            xaxis = xorigin + xscale * numpy.arange(data.shape[1])
            yaxis = yorigin + yscale * numpy.arange(data.shape[0])
            xlabel, ylabel = self._getAxesLabels(image)
            interpretation = "image" if len(data.shape) == 2 else "rgba-image"

            return save_NXdata(filename,
                               nxentry_name=entryPath,
                               signal=data,
                               axes=[yaxis, xaxis],
                               signal_name="image",
                               axes_names=["y", "x"],
                               axes_long_names=[ylabel, xlabel],
                               title=plot.getGraphTitle(),
                               interpretation=interpretation)

        elif nameFilter in (self.IMAGE_FILTER_ASCII,
                            self.IMAGE_FILTER_CSV_COMMA,
                            self.IMAGE_FILTER_CSV_SEMICOLON,
                            self.IMAGE_FILTER_CSV_TAB):
            csvdelim, filetype = {
                self.IMAGE_FILTER_ASCII: (' ', 'txt'),
                self.IMAGE_FILTER_CSV_COMMA: (',', 'csv'),
                self.IMAGE_FILTER_CSV_SEMICOLON: (';', 'csv'),
                self.IMAGE_FILTER_CSV_TAB: ('\t', 'csv'),
                }[nameFilter]

            height, width = data.shape
            rows, cols = numpy.mgrid[0:height, 0:width]
            try:
                save1D(filename, rows.ravel(), (cols.ravel(), data.ravel()),
                       filetype=filetype,
                       xlabel='row',
                       ylabels=['column', 'value'],
                       csvdelim=csvdelim,
                       autoheader=True)

            except IOError:
                self._errorMessage('Save failed\n', parent=self.plot)
                return False
            return True

        elif nameFilter == self.IMAGE_FILTER_RGB_PNG:
            # Get displayed image
            rgbaImage = image.getRgbaImageData(copy=False)
            # Convert RGB QImage
            qimage = convertArrayToQImage(rgbaImage[:, :, :3])

            if qimage.save(filename, 'PNG'):
                return True
            else:
                _logger.error('Failed to save image as %s', filename)
                qt.QMessageBox.critical(
                    self.parent(),
                    'Save image as',
                    'Failed to save image')

        return False

    def _saveScatter(self, plot, filename, nameFilter):
        """Save an image from the plot.

        :param str filename: The name of the file to write
        :param str nameFilter: The selected name filter
        :return: False if format is not supported or save failed,
                 True otherwise.
        """
        if nameFilter not in self.DEFAULT_SCATTER_FILTERS:
            return False

        if nameFilter == self.SCATTER_FILTER_NXDATA:
            entryPath = self._selectWriteableOutputGroup(filename, parent=self.plot)
            if entryPath is None:
                return False
            scatter = plot.getScatter()

            x = scatter.getXData(copy=False)
            y = scatter.getYData(copy=False)
            z = scatter.getValueData(copy=False)

            xerror = scatter.getXErrorData(copy=False)
            if isinstance(xerror, float):
                xerror = xerror * numpy.ones(x.shape, dtype=numpy.float32)

            yerror = scatter.getYErrorData(copy=False)
            if isinstance(yerror, float):
                yerror = yerror * numpy.ones(x.shape, dtype=numpy.float32)

            xlabel = plot.getGraphXLabel()
            ylabel = plot.getGraphYLabel()

            return save_NXdata(
                filename,
                nxentry_name=entryPath,
                signal=z,
                axes=[x, y],
                signal_name="values",
                axes_names=["x", "y"],
                axes_long_names=[xlabel, ylabel],
                axes_errors=[xerror, yerror],
                title=plot.getGraphTitle())

    def setFileFilter(self, dataKind, nameFilter, func, index=None, appendToFile=False):
        """Set a name filter to add/replace a file format support

        :param str dataKind:
            The kind of data for which the provided filter is valid.
            One of: 'all', 'curve', 'curves', 'image', 'scatter'
        :param str nameFilter: The name filter in the QFileDialog.
            See :meth:`QFileDialog.setNameFilters`.
        :param callable func: The function to call to perform saving.
            Expected signature is:
            bool func(PlotWidget plot, str filename, str nameFilter)
        :param bool appendToFile: True to append the data into the selected
            file.
        :param integer index: Index of the filter in the final list (or None)
        """
        assert dataKind in ('all', 'curve', 'curves', 'image', 'scatter')

        if appendToFile:
            self._appendFilters.append(nameFilter)

        # first append or replace the new filter to prevent colissions
        self._filters[dataKind][nameFilter] = func
        if index is None:
            # we are already done
            return

        # get the current ordered list of keys
        keyList = list(self._filters[dataKind].keys())

        # deal with negative indices
        if index < 0:
            index = len(keyList) + index
            if index < 0:
                index = 0

        if index >= len(keyList):
            # nothing to be done, already at the end
            txt = 'Requested index %d impossible, already at the end' % index
            _logger.info(txt)
            return

        # get the new ordered list
        oldIndex = keyList.index(nameFilter)
        del keyList[oldIndex]
        keyList.insert(index, nameFilter)

        # build the new filters
        newFilters = OrderedDict()
        for key in keyList:
            newFilters[key] = self._filters[dataKind][key]

        # and update the filters
        self._filters[dataKind] = newFilters
        return

    def getFileFilters(self, dataKind):
        """Returns the nameFilter and associated function for a kind of data.

        :param str dataKind:
            The kind of data for which the provided filter is valid.
            On of: 'all', 'curve', 'curves', 'image', 'scatter'
        :return: {nameFilter: function} associations.
        :rtype: collections.OrderedDict
        """
        assert dataKind in ('all', 'curve', 'curves', 'image', 'scatter')

        return self._filters[dataKind].copy()

    def _actionTriggered(self, checked=False):
        """Handle save action."""
        # Set-up filters
        filters = OrderedDict()

        # Add image filters if there is an active image
        if self.plot.getActiveImage() is not None:
            filters.update(self._filters['image'].items())

        # Add curve filters if there is a curve to save
        if (self.plot.getActiveCurve() is not None or
                len(self.plot.getAllCurves()) == 1):
            filters.update(self._filters['curve'].items())
        if len(self.plot.getAllCurves()) >= 1:
            filters.update(self._filters['curves'].items())

        # Add scatter filters if there is a scatter
        # todo: CSV
        if self.plot.getScatter() is not None:
            filters.update(self._filters['scatter'].items())

        filters.update(self._filters['all'].items())

        # Create and run File dialog
        dialog = qt.QFileDialog(self.plot)
        dialog.setOption(dialog.DontUseNativeDialog)
        dialog.setWindowTitle("Output File Selection")
        dialog.setModal(1)
        dialog.setNameFilters(list(filters.keys()))

        dialog.setFileMode(dialog.AnyFile)
        dialog.setAcceptMode(dialog.AcceptSave)

        def onFilterSelection(filt_):
            # disable overwrite confirmation for NXdata types,
            # because we append the data to existing files
            if filt_ in self._appendFilters:
                dialog.setOption(dialog.DontConfirmOverwrite)
            else:
                dialog.setOption(dialog.DontConfirmOverwrite, False)

        dialog.filterSelected.connect(onFilterSelection)

        if not dialog.exec_():
            return False

        nameFilter = dialog.selectedNameFilter()
        filename = dialog.selectedFiles()[0]
        dialog.close()

        if '(' in nameFilter and ')' == nameFilter.strip()[-1]:
            # Check for correct file extension
            # Extract file extensions as .something
            extensions = [ext[ext.find('.'):] for ext in
                          nameFilter[nameFilter.find('(') + 1:-1].split()]
            for ext in extensions:
                if (len(filename) > len(ext) and
                        filename[-len(ext):].lower() == ext.lower()):
                    break
            else:  # filename has no extension supported in nameFilter, add one
                if len(extensions) >= 1:
                    filename += extensions[0]

        # Handle save
        func = filters.get(nameFilter, None)
        if func is not None:
            return func(self.plot, filename, nameFilter)
        else:
            _logger.error('Unsupported file filter: %s', nameFilter)
            return False


def _plotAsPNG(plot):
    """Save a :class:`Plot` as PNG and return the payload.

    :param plot: The :class:`Plot` to save
    """
    pngFile = BytesIO()
    plot.saveGraph(pngFile, fileFormat='png')
    pngFile.flush()
    pngFile.seek(0)
    data = pngFile.read()
    pngFile.close()
    return data


class PrintAction(PlotAction):
    """QAction for printing the plot.

    It opens a Print dialog.

    Current implementation print a bitmap of the plot area and not vector
    graphics, so printing quality is not great.

    :param plot: :class:`.PlotWidget` instance on which to operate.
    :param parent: See :class:`QAction`.
    """

    def __init__(self, plot, parent=None):
        super(PrintAction, self).__init__(
            plot, icon='document-print', text='Print...',
            tooltip='Open print dialog',
            triggered=self.printPlot,
            checkable=False, parent=parent)
        self.setShortcut(qt.QKeySequence.Print)
        self.setShortcutContext(qt.Qt.WidgetShortcut)

    def getPrinter(self):
        """The QPrinter instance used by the PrintAction.

        :rtype: QPrinter
        """
        return printer.getDefaultPrinter()

    @property
    @deprecated(replacement="getPrinter()", since_version="0.8.0")
    def printer(self):
        return self.getPrinter()

    def printPlotAsWidget(self):
        """Open the print dialog and print the plot.

        Use :meth:`QWidget.render` to print the plot

        :return: True if successful
        """
        dialog = qt.QPrintDialog(self.getPrinter(), self.plot)
        dialog.setWindowTitle('Print Plot')
        if not dialog.exec_():
            return False

        # Print a snapshot of the plot widget at the top of the page
        widget = self.plot.centralWidget()

        painter = qt.QPainter()
        if not painter.begin(self.getPrinter()):
            return False

        pageRect = self.getPrinter().pageRect()
        xScale = pageRect.width() / widget.width()
        yScale = pageRect.height() / widget.height()
        scale = min(xScale, yScale)

        painter.translate(pageRect.width() / 2., 0.)
        painter.scale(scale, scale)
        painter.translate(-widget.width() / 2., 0.)
        widget.render(painter)
        painter.end()

        return True

    def printPlot(self):
        """Open the print dialog and print the plot.

        Use :meth:`Plot.saveGraph` to print the plot.

        :return: True if successful
        """
        # Init printer and start printer dialog
        dialog = qt.QPrintDialog(self.getPrinter(), self.plot)
        dialog.setWindowTitle('Print Plot')
        if not dialog.exec_():
            return False

        # Save Plot as PNG and make a pixmap from it with default dpi
        pngData = _plotAsPNG(self.plot)

        pixmap = qt.QPixmap()
        pixmap.loadFromData(pngData, 'png')

        xScale = self.getPrinter().pageRect().width() / pixmap.width()
        yScale = self.getPrinter().pageRect().height() / pixmap.height()
        scale = min(xScale, yScale)

        # Draw pixmap with painter
        painter = qt.QPainter()
        if not painter.begin(self.getPrinter()):
            return False

        painter.drawPixmap(0, 0,
                           pixmap.width() * scale,
                           pixmap.height() * scale,
                           pixmap)
        painter.end()

        return True


class CopyAction(PlotAction):
    """QAction to copy :class:`.PlotWidget` content to clipboard.

    :param plot: :class:`.PlotWidget` instance on which to operate
    :param parent: See :class:`QAction`
    """

    def __init__(self, plot, parent=None):
        super(CopyAction, self).__init__(
            plot, icon='edit-copy', text='Copy plot',
            tooltip='Copy a snapshot of the plot into the clipboard',
            triggered=self.copyPlot,
            checkable=False, parent=parent)
        self.setShortcut(qt.QKeySequence.Copy)
        self.setShortcutContext(qt.Qt.WidgetShortcut)

    def copyPlot(self):
        """Copy plot content to the clipboard as a bitmap."""
        # Save Plot as PNG and make a QImage from it with default dpi
        pngData = _plotAsPNG(self.plot)
        image = qt.QImage.fromData(pngData, 'png')
        qt.QApplication.clipboard().setImage(image)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module provides a widget displaying mouse coordinates in a PlotWidget.

It can be configured to provide more information.
"""

from __future__ import division

__authors__ = ["V.A. Sole", "T. Vincent"]
__license__ = "MIT"
__date__ = "16/10/2017"


import logging
import numbers
import traceback
import weakref

import numpy

from ....utils.deprecation import deprecated
from ... import qt
from .. import items


_logger = logging.getLogger(__name__)


# PositionInfo ################################################################

class PositionInfo(qt.QWidget):
    """QWidget displaying coords converted from data coords of the mouse.

    Provide this widget with a list of couple:

    - A name to display before the data
    - A function that takes (x, y) as arguments and returns something that
      gets converted to a string.
      If the result is a float it is converted with '%.7g' format.

    To run the following sample code, a QApplication must be initialized.
    First, create a PlotWindow and add a QToolBar where to place the
    PositionInfo widget.

    >>> from silx.gui.plot import PlotWindow
    >>> from silx.gui import qt

    >>> plot = PlotWindow()  # Create a PlotWindow to add the widget to
    >>> toolBar = qt.QToolBar()  # Create a toolbar to place the widget in
    >>> plot.addToolBar(qt.Qt.BottomToolBarArea, toolBar)  # Add it to plot

    Then, create the PositionInfo widget and add it to the toolbar.
    The PositionInfo widget is created with a list of converters, here
    to display polar coordinates of the mouse position.

    >>> import numpy
    >>> from silx.gui.plot.tools import PositionInfo

    >>> position = PositionInfo(plot=plot, converters=[
    ...     ('Radius', lambda x, y: numpy.sqrt(x*x + y*y)),
    ...     ('Angle', lambda x, y: numpy.degrees(numpy.arctan2(y, x)))])
    >>> toolBar.addWidget(position)  # Add the widget to the toolbar
    <...>
    >>> plot.show()  # To display the PlotWindow with the position widget

    :param plot: The PlotWidget this widget is displaying data coords from.
    :param converters:
        List of 2-tuple: name to display and conversion function from (x, y)
        in data coords to displayed value.
        If None, the default, it displays X and Y.
    :param parent: Parent widget
    """

    SNAP_THRESHOLD_DIST = 5

    def __init__(self, parent=None, plot=None, converters=None):
        assert plot is not None
        self._plotRef = weakref.ref(plot)
        self._snappingMode = self.SNAPPING_DISABLED

        super(PositionInfo, self).__init__(parent)

        if converters is None:
            converters = (('X', lambda x, y: x), ('Y', lambda x, y: y))

        self._fields = []  # To store (QLineEdit, name, function (x, y)->v)

        # Create a new layout with new widgets
        layout = qt.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        # layout.setSpacing(0)

        # Create all QLabel and store them with the corresponding converter
        for name, func in converters:
            layout.addWidget(qt.QLabel('<b>' + name + ':</b>'))

            contentWidget = qt.QLabel()
            contentWidget.setText('------')
            contentWidget.setTextInteractionFlags(qt.Qt.TextSelectableByMouse)
            contentWidget.setFixedWidth(
                contentWidget.fontMetrics().boundingRect('##############').width())
            layout.addWidget(contentWidget)
            self._fields.append((contentWidget, name, func))

        layout.addStretch(1)
        self.setLayout(layout)

        # Connect to Plot events
        plot.sigPlotSignal.connect(self._plotEvent)

    def getPlotWidget(self):
        """Returns the PlotWidget this widget is attached to or None.

        :rtype: Union[~silx.gui.plot.PlotWidget,None]
        """
        return self._plotRef()

    @property
    @deprecated(replacement='getPlotWidget', since_version='0.8.0')
    def plot(self):
        return self.getPlotWidget()

    def getConverters(self):
        """Return the list of converters as 2-tuple (name, function)."""
        return [(name, func) for _label, name, func in self._fields]

    def _plotEvent(self, event):
        """Handle events from the Plot.

        :param dict event: Plot event
        """
        if event['event'] == 'mouseMoved':
            x, y = event['x'], event['y']
            xPixel, yPixel = event['xpixel'], event['ypixel']
            self._updateStatusBar(x, y, xPixel, yPixel)

    def updateInfo(self):
        """Update displayed information"""
        plot = self.getPlotWidget()
        if plot is None:
            _logger.error("Trying to update PositionInfo "
                          "while PlotWidget no longer exists")
            return

        widget = plot.getWidgetHandle()
        position = widget.mapFromGlobal(qt.QCursor.pos())
        xPixel, yPixel = position.x(), position.y()
        dataPos = plot.pixelToData(xPixel, yPixel, check=True)
        if dataPos is not None:  # Inside plot area
            x, y = dataPos
            self._updateStatusBar(x, y, xPixel, yPixel)

    def _updateStatusBar(self, x, y, xPixel, yPixel):
        """Update information from the status bar using the definitions.

        :param float x: Position-x in data
        :param float y: Position-y in data
        :param float xPixel: Position-x in pixels
        :param float yPixel: Position-y in pixels
        """
        plot = self.getPlotWidget()
        if plot is None:
            return

        styleSheet = "color: rgb(0, 0, 0);"  # Default style
        xData, yData = x, y

        snappingMode = self.getSnappingMode()

        # Snapping when crosshair either not requested or active
        if (snappingMode & (self.SNAPPING_CURVE | self.SNAPPING_SCATTER) and
                (not (snappingMode & self.SNAPPING_CROSSHAIR) or
                 plot.getGraphCursor())):
            styleSheet = "color: rgb(255, 0, 0);"  # Style far from item

            if snappingMode & self.SNAPPING_ACTIVE_ONLY:
                selectedItems = []

                if snappingMode & self.SNAPPING_CURVE:
                    activeCurve = plot.getActiveCurve()
                    if activeCurve:
                        selectedItems.append(activeCurve)

                if snappingMode & self.SNAPPING_SCATTER:
                    activeScatter = plot._getActiveItem(kind='scatter')
                    if activeScatter:
                        selectedItems.append(activeScatter)

            else:
                kinds = []
                if snappingMode & self.SNAPPING_CURVE:
                    kinds.append(items.Curve)
                if snappingMode & self.SNAPPING_SCATTER:
                    kinds.append(items.Scatter)
                selectedItems = [item for item in plot.getItems()
                                 if isinstance(item, kinds) and item.isVisible()]

            # Compute distance threshold
            if qt.BINDING in ('PyQt5', 'PySide2'):
                window = plot.window()
                windowHandle = window.windowHandle()
                if windowHandle is not None:
                    ratio = windowHandle.devicePixelRatio()
                else:
                    ratio = qt.QGuiApplication.primaryScreen().devicePixelRatio()
            else:
                ratio = 1.

            # Baseline squared distance threshold
            distInPixels = (self.SNAP_THRESHOLD_DIST * ratio)**2

            for item in selectedItems:
                if (snappingMode & self.SNAPPING_SYMBOLS_ONLY and
                        not item.getSymbol()):
                    # Only handled if item symbols are visible
                    continue

                xArray = item.getXData(copy=False)
                yArray = item.getYData(copy=False)
                closestIndex = numpy.argmin(
                    pow(xArray - x, 2) + pow(yArray - y, 2))

                xClosest = xArray[closestIndex]
                yClosest = yArray[closestIndex]

                if isinstance(item, items.YAxisMixIn):
                    axis = item.getYAxis()
                else:
                    axis = 'left'

                closestInPixels = plot.dataToPixel(
                    xClosest, yClosest, axis=axis)
                if closestInPixels is not None:
                    curveDistInPixels = (
                        (closestInPixels[0] - xPixel)**2 +
                        (closestInPixels[1] - yPixel)**2)

                    if curveDistInPixels <= distInPixels:
                        # Update label style sheet
                        styleSheet = "color: rgb(0, 0, 0);"

                        # if close enough, snap to data point coord
                        xData, yData = xClosest, yClosest
                        distInPixels = curveDistInPixels

        for label, name, func in self._fields:
            label.setStyleSheet(styleSheet)

            try:
                value = func(xData, yData)
                text = self.valueToString(value)
                label.setText(text)
            except:
                label.setText('Error')
                _logger.error(
                    "Error while converting coordinates (%f, %f)"
                    "with converter '%s'" % (xPixel, yPixel, name))
                _logger.error(traceback.format_exc())

    def valueToString(self, value):
        if isinstance(value, (tuple, list)):
            value = [self.valueToString(v) for v in value]
            return ", ".join(value)
        elif isinstance(value, numbers.Real):
            # Use this for floats and int
            return '%.7g' % value
        else:
            # Fallback for other types
            return str(value)

    # Snapping mode

    SNAPPING_DISABLED = 0
    """No snapping occurs"""

    SNAPPING_CROSSHAIR = 1 << 0
    """Snapping only enabled when crosshair cursor is enabled"""

    SNAPPING_ACTIVE_ONLY = 1 << 1
    """Snapping only enabled for active item"""

    SNAPPING_SYMBOLS_ONLY = 1 << 2
    """Snapping only when symbols are visible"""

    SNAPPING_CURVE = 1 << 3
    """Snapping on curves"""

    SNAPPING_SCATTER = 1 << 4
    """Snapping on scatter"""

    def setSnappingMode(self, mode):
        """Set the snapping mode.

        The mode is a mask.

        :param int mode: The mode to use
        """
        if mode != self._snappingMode:
            self._snappingMode = mode
            self.updateInfo()

    def getSnappingMode(self):
        """Returns the snapping mode as a mask

        :rtype: int
        """
        return self._snappingMode

    _SNAPPING_LEGACY = (SNAPPING_CROSSHAIR |
                        SNAPPING_ACTIVE_ONLY |
                        SNAPPING_SYMBOLS_ONLY |
                        SNAPPING_CURVE |
                        SNAPPING_SCATTER)
    """Legacy snapping mode"""

    @property
    @deprecated(replacement="getSnappingMode", since_version="0.8")
    def autoSnapToActiveCurve(self):
        return self.getSnappingMode() == self._SNAPPING_LEGACY

    @autoSnapToActiveCurve.setter
    @deprecated(replacement="setSnappingMode", since_version="0.8")
    def autoSnapToActiveCurve(self, flag):
        self.setSnappingMode(
            self._SNAPPING_LEGACY if flag else self.SNAPPING_DISABLED)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Basic tests for StackView"""

__authors__ = ["P. Knobel"]
__license__ = "MIT"
__date__ = "20/03/2017"


import unittest
import numpy

from silx.gui.utils.testutils import TestCaseQt, SignalListener

from silx.gui import qt
from silx.gui.plot import StackView
from silx.gui.plot.StackView import StackViewMainWindow

from silx.utils.array_like import ListOfImages


class TestStackView(TestCaseQt):
    """Base class for tests of StackView."""

    def setUp(self):
        super(TestStackView, self).setUp()
        self.stackview = StackView()
        self.stackview.show()
        self.qWaitForWindowExposed(self.stackview)
        self.mystack = numpy.fromfunction(
            lambda i, j, k: numpy.sin(i/15.) + numpy.cos(j/4.) + 2 * numpy.sin(k/6.),
            (10, 20, 30)
        )

    def tearDown(self):
        self.stackview.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.stackview.close()
        del self.stackview
        super(TestStackView, self).tearDown()

    def testScaleColormapRangeToStack(self):
        """Test scaleColormapRangeToStack"""
        self.stackview.setStack(self.mystack)
        self.stackview.setColormap("viridis")
        colormap = self.stackview.getColormap()

        # Colormap autoscale to image
        self.assertEqual(colormap.getVRange(), (None, None))
        self.stackview.scaleColormapRangeToStack()

        # Colormap range set according to stack range
        self.assertEqual(colormap.getVRange(), (self.mystack.min(), self.mystack.max()))

    def testSetStack(self):
        self.stackview.setStack(self.mystack)
        self.stackview.setColormap("viridis", autoscale=True)
        my_trans_stack, params = self.stackview.getStack()
        self.assertEqual(my_trans_stack.shape, self.mystack.shape)
        self.assertTrue(numpy.array_equal(self.mystack,
                                          my_trans_stack))
        self.assertEqual(params["colormap"]["name"],
                         "viridis")

    def testSetStackPerspective(self):
        self.stackview.setStack(self.mystack, perspective=1)
        # my_orig_stack, params = self.stackview.getStack()
        my_trans_stack, params = self.stackview.getCurrentView()

        # get stack returns the transposed data, depending on the perspective
        self.assertEqual(my_trans_stack.shape,
                         (self.mystack.shape[1], self.mystack.shape[0], self.mystack.shape[2]))
        self.assertTrue(numpy.array_equal(numpy.transpose(self.mystack, axes=(1, 0, 2)),
                                          my_trans_stack))

    def testSetStackListOfImages(self):
        loi = [self.mystack[i] for i in range(self.mystack.shape[0])]

        self.stackview.setStack(loi)
        my_orig_stack, params = self.stackview.getStack(returnNumpyArray=True)
        my_trans_stack, params = self.stackview.getStack(returnNumpyArray=True)
        self.assertEqual(my_trans_stack.shape, self.mystack.shape)
        self.assertTrue(numpy.array_equal(self.mystack,
                                          my_trans_stack))
        self.assertTrue(numpy.array_equal(self.mystack,
                                          my_orig_stack))
        self.assertIsInstance(my_trans_stack, numpy.ndarray)

        self.stackview.setStack(loi, perspective=2)
        my_orig_stack, params = self.stackview.getStack(copy=False)
        my_trans_stack, params = self.stackview.getCurrentView(copy=False)
        # getStack(copy=False) must return the object set in setStack
        self.assertIs(my_orig_stack, loi)
        # getCurrentView(copy=False) returns a ListOfImages whose .images
        # attr is the original data
        self.assertEqual(my_trans_stack.shape,
                         (self.mystack.shape[2], self.mystack.shape[0], self.mystack.shape[1]))
        self.assertTrue(numpy.array_equal(numpy.array(my_trans_stack),
                                          numpy.transpose(self.mystack, axes=(2, 0, 1))))
        self.assertIsInstance(my_trans_stack,
                              ListOfImages)  # returnNumpyArray=False by default in getStack
        self.assertIs(my_trans_stack.images, loi)

    def testPerspective(self):
        self.stackview.setStack(numpy.arange(24).reshape((2, 3, 4)))
        self.assertEqual(self.stackview._perspective, 0,
                         "Default perspective is not 0 (dim1-dim2).")

        self.stackview._StackView__planeSelection.setPerspective(1)
        self.assertEqual(self.stackview._perspective, 1,
                         "Plane selection combobox not updating perspective")

        self.stackview.setStack(numpy.arange(6).reshape((1, 2, 3)))
        self.assertEqual(self.stackview._perspective, 1,
                         "Perspective not preserved when calling setStack "
                         "without specifying the perspective parameter.")

        self.stackview.setStack(numpy.arange(24).reshape((2, 3, 4)), perspective=2)
        self.assertEqual(self.stackview._perspective, 2,
                         "Perspective not set in setStack(..., perspective=2).")

    def testDefaultTitle(self):
        """Test that the plot title contains the proper Z information"""
        self.stackview.setStack(numpy.arange(24).reshape((4, 3, 2)),
                                calibrations=[(0, 1), (-10, 10), (3.14, 3.14)])
        self.assertEqual(self.stackview._plot.getGraphTitle(),
                         "Image z=0")
        self.stackview.setFrameNumber(2)
        self.assertEqual(self.stackview._plot.getGraphTitle(),
                         "Image z=2")

        self.stackview._StackView__planeSelection.setPerspective(1)
        self.stackview.setFrameNumber(0)
        self.assertEqual(self.stackview._plot.getGraphTitle(),
                         "Image z=-10")
        self.stackview.setFrameNumber(2)
        self.assertEqual(self.stackview._plot.getGraphTitle(),
                         "Image z=10")

        self.stackview._StackView__planeSelection.setPerspective(2)
        self.stackview.setFrameNumber(0)
        self.assertEqual(self.stackview._plot.getGraphTitle(),
                         "Image z=3.14")
        self.stackview.setFrameNumber(1)
        self.assertEqual(self.stackview._plot.getGraphTitle(),
                         "Image z=6.28")

    def testCustomTitle(self):
        """Test setting the plot title with a user defined callback"""
        self.stackview.setStack(numpy.arange(24).reshape((4, 3, 2)),
                                calibrations=[(0, 1), (-10, 10), (3.14, 3.14)])

        def title_callback(frame_idx):
            return "Cubed index title %d" % (frame_idx**3)

        self.stackview.setTitleCallback(title_callback)
        self.assertEqual(self.stackview._plot.getGraphTitle(),
                         "Cubed index title 0")
        self.stackview.setFrameNumber(2)
        self.assertEqual(self.stackview._plot.getGraphTitle(),
                         "Cubed index title 8")

        # perspective should not matter, only frame index
        self.stackview._StackView__planeSelection.setPerspective(1)
        self.stackview.setFrameNumber(0)
        self.assertEqual(self.stackview._plot.getGraphTitle(),
                         "Cubed index title 0")
        self.stackview.setFrameNumber(2)
        self.assertEqual(self.stackview._plot.getGraphTitle(),
                         "Cubed index title 8")

        with self.assertRaises(TypeError):
            # setTitleCallback should not accept non-callable objects like strings
            self.stackview.setTitleCallback(
                    "Là, vous faites sirop de vingt-et-un et vous dites : "
                    "beau sirop, mi-sirop, siroté, gagne-sirop, sirop-grelot,"
                    " passe-montagne, sirop au bon goût.")

    def testStackFrameNumber(self):
        self.stackview.setStack(self.mystack)
        self.assertEqual(self.stackview.getFrameNumber(), 0)

        listener = SignalListener()
        self.stackview.sigFrameChanged.connect(listener)

        self.stackview.setFrameNumber(1)
        self.assertEqual(self.stackview.getFrameNumber(), 1)
        self.assertEqual(listener.arguments(), [(1,)])


class TestStackViewMainWindow(TestCaseQt):
    """Base class for tests of StackView."""

    def setUp(self):
        super(TestStackViewMainWindow, self).setUp()
        self.stackview = StackViewMainWindow()
        self.stackview.show()
        self.qWaitForWindowExposed(self.stackview)
        self.mystack = numpy.fromfunction(
            lambda i, j, k: numpy.sin(i/15.) + numpy.cos(j/4.) + 2 * numpy.sin(k/6.),
            (10, 20, 30)
        )

    def tearDown(self):
        self.stackview.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.stackview.close()
        del self.stackview
        super(TestStackViewMainWindow, self).tearDown()

    def testSetStack(self):
        self.stackview.setStack(self.mystack)
        self.stackview.setColormap("viridis", autoscale=True)
        my_trans_stack, params = self.stackview.getStack()
        self.assertEqual(my_trans_stack.shape, self.mystack.shape)
        self.assertTrue(numpy.array_equal(self.mystack,
                                          my_trans_stack))
        self.assertEqual(params["colormap"]["name"],
                         "viridis")

    def testSetStackPerspective(self):
        self.stackview.setStack(self.mystack, perspective=1)
        my_trans_stack, params = self.stackview.getCurrentView()
        # get stack returns the transposed data, depending on the perspective
        self.assertEqual(my_trans_stack.shape,
                         (self.mystack.shape[1], self.mystack.shape[0], self.mystack.shape[2]))
        self.assertTrue(numpy.array_equal(numpy.transpose(self.mystack, axes=(1, 0, 2)),
                                          my_trans_stack))


def suite():
    test_suite = unittest.TestSuite()
    test_suite.addTest(
        unittest.defaultTestLoader.loadTestsFromTestCase(TestStackView))
    test_suite.addTest(
        unittest.defaultTestLoader.loadTestsFromTestCase(TestStackViewMainWindow))
    return test_suite


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# coding: utf-8
#/*##########################################################################
#
# Copyright (c) 2004-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
########################################################################### */
"""
This module defines the :class:`FitTheory` object that is used by
:class:`silx.math.fit.FitManager` to define fit functions and background
models.
"""

__authors__ = ["P. Knobel"]
__license__ = "MIT"
__date__ = "09/08/2016"


class FitTheory(object):
    """This class defines a fit theory, which consists of:

        - a model function, the actual function to be fitted
        - parameters names
        - an estimation function, that return the estimated initial parameters
          that serve as input for :func:`silx.math.fit.leastsq`
        - an optional configuration function, that can be used to modify
          configuration parameters to alter the behavior of the fit function
          and the estimation function
        - an optional derivative function, that replaces the default model
          derivative used in :func:`silx.math.fit.leastsq`
    """
    def __init__(self, function, parameters,
                 estimate=None, configure=None, derivative=None,
                 description=None, pymca_legacy=False, is_background=False):
        """
        :param function function: Actual function. See documentation for
            :attr:`function`.
        :param list[str] parameters: List of parameter names for the function.
            See documentation for :attr:`parameters`.
        :param function estimate: Optional estimation function.
            See documentation for :attr:`estimate`
        :param function configure: Optional configuration function.
            See documentation for :attr:`configure`
        :param function derivative: Optional custom derivative function.
            See documentation for :attr:`derivative`
        :param str description: Optional description string.
            See documentation for :attr:`description`
        :param bool pymca_legacy: Flag to indicate that the theory is a PyMca
            legacy theory. See documentation for :attr:`pymca_legacy`
        :param bool is_background: Flag to indicate that the theory is a
            background theory. This has implications regarding the function's
            signature, as explained in the documentation for :attr:`function`.
        """
        self.function = function
        """Regular fit functions must have the signature ``f(x, *params) -> y``,
        where *x* is a 1D array of values for the independent variable,
        *params* are the parameters to be fitted and *y* is the output array
        that we want to have the best fit to a series of data points.

        Background functions used by :class:`FitManager` must have a slightly
        different signature: ``f(x, y0, *params) -> bg``, where *y0* is the
        array of original data points and *bg* is the background signal that
        we want to subtract from the data array prior to fitting the regular
        fit function.

        The number of parameters must be the same as in :attr:`parameters`, or
        a multiple of this number if the function is defined as a sum of a
        variable number of base functions and if :attr:`estimate` is designed
        to be able to estimate the number of needed base functions.
        """

        self.parameters = parameters
        """List of parameters names.

        This list can contain the minimum number of parameters, if the
        function takes a variable number of parameters,
        and if the estimation function is responsible for finding the number
        of required parameters """

        self.estimate = estimate
        """The estimation function should have the following signature::

            f(x, y) -> (estimated_param, constraints)

        Parameters:

            - ``x`` is a sequence of values for the independent variable
            - ``y`` is a sequence of the same length as ``x`` containing the
              data to be fitted

        Return values:

            - ``estimated_param`` is a sequence of estimated fit parameters to
              be used as initial values for an iterative fit.
            - ``constraints`` is a sequence of shape *(n, 3)*, where *n* is the
              number of estimated parameters, containing the constraints for each
              parameter to be fitted. See :func:`silx.math.fit.leastsq` for more
              explanations about constraints."""
        if estimate is None:
            self.estimate = self.default_estimate

        self.configure = configure
        """The optional configuration function must conform to the signature
        ``f(**kw) -> dict`` (i.e it must accept any named argument and
        return a dictionary).
        It can be used to modify configuration parameters to alter the
        behavior of the fit function and the estimation function."""

        self.derivative = derivative
        """The optional derivative function must conform to the signature
        ``model_deriv(xdata, parameters, index)``, where parameters is a
        sequence with the current values of the fitting parameters, index is
        the fitting parameter index for which the the derivative has to be
        provided in the supplied array of xdata points."""

        self.description = description
        """Optional description string for this particular fit theory."""

        self.pymca_legacy = pymca_legacy
        """This attribute can be set to *True* to indicate that the theory
        is a PyMca legacy theory.

        This tells :mod:`silx.math.fit.fitmanager` that the signature of
        the estimate function is::

            f(x, y, bg, xscaling, yscaling) -> (estimated_param, constraints)
        """

        self.is_background = is_background
        """Flag to indicate that the theory is background theory.

        A background function is an secondary function that needs to be added
        to the main fit function to better fit the original data.
        If this flag is set to *True*, modules using this theory are informed
        that :attr:`function` has the signature ``f(x, y0, *params) -> bg``,
        instead of the usual fit function signature."""

    def default_estimate(self, x=None, y=None, bg=None):
        """Default estimate function. Return an array of *ones* as the
        initial estimated parameters, and set all constraints to zero
        (FREE)"""
        estimated_parameters = [1. for _ in self.parameters]
        estimated_constraints = [[0, 0, 0] for _ in self.parameters]
        return estimated_parameters, estimated_constraints



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2004-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Functions to apply pan and zoom on a Plot"""

__authors__ = ["T. Vincent", "V. Valls"]
__license__ = "MIT"
__date__ = "08/08/2017"


import math
import numpy


# Float 32 info ###############################################################
# Using min/max value below limits of float32
# so operation with such value (e.g., max - min) do not overflow

FLOAT32_SAFE_MIN = -1e37
FLOAT32_MINPOS = numpy.finfo(numpy.float32).tiny
FLOAT32_SAFE_MAX = 1e37
# TODO double support


def scale1DRange(min_, max_, center, scale, isLog):
    """Scale a 1D range given a scale factor and an center point.

    Keeps the values in a smaller range than float32.

    :param float min_: The current min value of the range.
    :param float max_: The current max value of the range.
    :param float center: The center of the zoom (i.e., invariant point).
    :param float scale: The scale to use for zoom
    :param bool isLog: Whether using log scale or not.
    :return: The zoomed range.
    :rtype: tuple of 2 floats: (min, max)
    """
    if isLog:
        # Min and center can be < 0 when
        # autoscale is off and switch to log scale
        # max_ < 0 should not happen
        min_ = numpy.log10(min_) if min_ > 0. else FLOAT32_MINPOS
        center = numpy.log10(center) if center > 0. else FLOAT32_MINPOS
        max_ = numpy.log10(max_) if max_ > 0. else FLOAT32_MINPOS

    if min_ == max_:
        return min_, max_

    offset = (center - min_) / (max_ - min_)
    range_ = (max_ - min_) / scale
    newMin = center - offset * range_
    newMax = center + (1. - offset) * range_

    if isLog:
        # No overflow as exponent is log10 of a float32
        newMin = pow(10., newMin)
        newMax = pow(10., newMax)
        newMin = numpy.clip(newMin, FLOAT32_MINPOS, FLOAT32_SAFE_MAX)
        newMax = numpy.clip(newMax, FLOAT32_MINPOS, FLOAT32_SAFE_MAX)
    else:
        newMin = numpy.clip(newMin, FLOAT32_SAFE_MIN, FLOAT32_SAFE_MAX)
        newMax = numpy.clip(newMax, FLOAT32_SAFE_MIN, FLOAT32_SAFE_MAX)
    return newMin, newMax


def applyZoomToPlot(plot, scaleF, center=None):
    """Zoom in/out plot given a scale and a center point.

    :param plot: The plot on which to apply zoom.
    :param float scaleF: Scale factor of zoom.
    :param center: (x, y) coords in pixel coordinates of the zoom center.
    :type center: 2-tuple of float
    """
    xMin, xMax = plot.getXAxis().getLimits()
    yMin, yMax = plot.getYAxis().getLimits()

    if center is None:
        left, top, width, height = plot.getPlotBoundsInPixels()
        cx, cy = left + width // 2, top + height // 2
    else:
        cx, cy = center

    dataCenterPos = plot.pixelToData(cx, cy)
    assert dataCenterPos is not None

    xMin, xMax = scale1DRange(xMin, xMax, dataCenterPos[0], scaleF,
                              plot.getXAxis()._isLogarithmic())

    yMin, yMax = scale1DRange(yMin, yMax, dataCenterPos[1], scaleF,
                              plot.getYAxis()._isLogarithmic())

    dataPos = plot.pixelToData(cx, cy, axis="right")
    assert dataPos is not None
    y2Center = dataPos[1]
    y2Min, y2Max = plot.getYAxis(axis="right").getLimits()
    y2Min, y2Max = scale1DRange(y2Min, y2Max, y2Center, scaleF,
                                plot.getYAxis()._isLogarithmic())

    plot.setLimits(xMin, xMax, yMin, yMax, y2Min, y2Max)


def applyPan(min_, max_, panFactor, isLog10):
    """Returns a new range with applied panning.

    Moves the range according to panFactor.
    If isLog10 is True, converts to log10 before moving.

    :param float min_: Min value of the data range to pan.
    :param float max_: Max value of the data range to pan.
                       Must be >= min.
    :param float panFactor: Signed proportion of the range to use for pan.
    :param bool isLog10: True if log10 scale, False if linear scale.
    :return: New min and max value with pan applied.
    :rtype: 2-tuple of float.
    """
    if isLog10 and min_ > 0.:
        # Negative range and log scale can happen with matplotlib
        logMin, logMax = math.log10(min_), math.log10(max_)
        logOffset = panFactor * (logMax - logMin)
        newMin = pow(10., logMin + logOffset)
        newMax = pow(10., logMax + logOffset)

        # Takes care of out-of-range values
        if newMin > 0. and newMax < float('inf'):
            min_, max_ = newMin, newMax

    else:
        offset = panFactor * (max_ - min_)
        newMin, newMax = min_ + offset, max_ + offset

        # Takes care of out-of-range values
        if newMin > - float('inf') and newMax < float('inf'):
            min_, max_ = newMin, newMax
    return min_, max_


class _Unset(object):
    """To be able to have distinction between None and unset"""
    pass


class ViewConstraints(object):
    """
    Store constraints applied on the view box and compute the resulting view box.
    """

    def __init__(self):
        self._min = [None, None]
        self._max = [None, None]
        self._minRange = [None, None]
        self._maxRange = [None, None]

    def update(self, xMin=_Unset, xMax=_Unset,
               yMin=_Unset, yMax=_Unset,
               minXRange=_Unset, maxXRange=_Unset,
               minYRange=_Unset, maxYRange=_Unset):
        """
        Update the constraints managed by the object

        The constraints are the same as the ones provided by PyQtGraph.

        :param float xMin: Minimum allowed x-axis value.
            (default do not change the stat, None remove the constraint)
        :param float xMax: Maximum allowed x-axis value.
            (default do not change the stat, None remove the constraint)
        :param float yMin: Minimum allowed y-axis value.
            (default do not change the stat, None remove the constraint)
        :param float yMax: Maximum allowed y-axis value.
            (default do not change the stat, None remove the constraint)
        :param float minXRange: Minimum allowed left-to-right span across the
            view (default do not change the stat, None remove the constraint)
        :param float maxXRange: Maximum allowed left-to-right span across the
            view (default do not change the stat, None remove the constraint)
        :param float minYRange: Minimum allowed top-to-bottom span across the
            view (default do not change the stat, None remove the constraint)
        :param float maxYRange: Maximum allowed top-to-bottom span across the
            view (default do not change the stat, None remove the constraint)
        :return: True if the constraints was changed
        """
        updated = False

        minRange = [minXRange, minYRange]
        maxRange = [maxXRange, maxYRange]
        minPos = [xMin, yMin]
        maxPos = [xMax, yMax]

        for axis in range(2):

            value = minPos[axis]
            if value is not _Unset and value != self._min[axis]:
                self._min[axis] = value
                updated = True

            value = maxPos[axis]
            if value is not _Unset and value != self._max[axis]:
                self._max[axis] = value
                updated = True

            value = minRange[axis]
            if value is not _Unset and value != self._minRange[axis]:
                self._minRange[axis] = value
                updated = True

            value = maxRange[axis]
            if value is not _Unset and value != self._maxRange[axis]:
                self._maxRange[axis] = value
                updated = True

        # Sanity checks

        for axis in range(2):
            if self._maxRange[axis] is not None and self._min[axis] is not None and self._max[axis] is not None:
                # max range cannot be larger than bounds
                diff = self._max[axis] - self._min[axis]
                self._maxRange[axis] = min(self._maxRange[axis], diff)
                updated = True

        return updated

    def normalize(self, xMin, xMax, yMin, yMax, allow_scaling=True):
        """Normalize a view range defined by x and y corners using predefined
        containts.

        :param float xMin: Min position of the x-axis
        :param float xMax: Max position of the x-axis
        :param float yMin: Min position of the y-axis
        :param float yMax: Max position of the y-axis
        :param bool allow_scaling: Allow or not to apply scaling for the
            normalization. Used according to the interaction mode.
        :return: A normalized tuple of (xMin, xMax, yMin, yMax)
        """
        viewRange = [[xMin, xMax], [yMin, yMax]]

        for axis in range(2):
            # clamp xRange and yRange
            if allow_scaling:
                diff = viewRange[axis][1] - viewRange[axis][0]
                delta = None
                if self._maxRange[axis] is not None and diff > self._maxRange[axis]:
                    delta = self._maxRange[axis] - diff
                elif self._minRange[axis] is not None and diff < self._minRange[axis]:
                    delta = self._minRange[axis] - diff
                if delta is not None:
                    viewRange[axis][0] -= delta * 0.5
                    viewRange[axis][1] += delta * 0.5

            # clamp min and max positions
            outMin = self._min[axis] is not None and viewRange[axis][0] < self._min[axis]
            outMax = self._max[axis] is not None and viewRange[axis][1] > self._max[axis]

            if outMin and outMax:
                if allow_scaling:
                    # we can clamp both sides
                    viewRange[axis][0] = self._min[axis]
                    viewRange[axis][1] = self._max[axis]
                else:
                    # center the result
                    delta = viewRange[axis][1] - viewRange[axis][0]
                    mid = self._min[axis] + self._max[axis] - self._min[axis]
                    viewRange[axis][0] = mid - delta
                    viewRange[axis][1] = mid + delta
            elif outMin:
                delta = self._min[axis] - viewRange[axis][0]
                viewRange[axis][0] += delta
                viewRange[axis][1] += delta
            elif outMax:
                delta = self._max[axis] - viewRange[axis][1]
                viewRange[axis][0] += delta
                viewRange[axis][1] += delta

        return viewRange[0][0], viewRange[0][1], viewRange[1][0], viewRange[1][1]



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Utils function relative to files
"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "19/09/2016"

import os.path
import glob

def expand_filenames(filenames):
    """
    Takes a list of paths and expand it into a list of files.

    :param List[str] filenames: list of filenames or path with wildcards
    :rtype: List[str]
    :return: list of existing filenames or non-existing files
        (which was provided as input)
    """
    result = []
    for filename in filenames:
        if os.path.exists(filename):
            result.append(filename)
        elif glob.has_magic(filename):
            expanded_filenames = glob.glob(filename)
            if expanded_filenames:
                result += expanded_filenames
            else:  # Cannot expand, add as is
                result.append(filename)
        else:
            result.append(filename)
    return result



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2015-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

from __future__ import absolute_import, division, unicode_literals

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "17/01/2018"


import unittest
from silx.utils.testutils import ParametricTestCase

import numpy

from silx.gui.plot3d.scene import utils


# angleBetweenVectors #########################################################

class TestAngleBetweenVectors(ParametricTestCase):

    TESTS = {  # name: (refvector, vectors, norm, refangles)
        'single vector':
            ((1., 0., 0.), (1., 0., 0.), (0., 0., 1.), 0.),
        'single vector, no norm':
            ((1., 0., 0.), (1., 0., 0.), None, 0.),

        'with orthogonal norm':
            ((1., 0., 0.),
             ((1., 0., 0.), (0., 1., 0.), (-1., 0., 0.), (0., -1., 0.)),
             (0., 0., 1.),
             (0., 90., 180., 270.)),

        'with coplanar norm':  # = similar to no norm
            ((1., 0., 0.),
             ((1., 0., 0.), (0., 1., 0.), (-1., 0., 0.), (0., -1., 0.)),
             (1., 0., 0.),
             (0., 90., 180., 90.)),

        'without norm':
            ((1., 0., 0.),
             ((1., 0., 0.), (0., 1., 0.), (-1., 0., 0.), (0., -1., 0.)),
             None,
             (0., 90., 180., 90.)),

        'not unit vectors':
            ((2., 2., 0.), ((1., 1., 0.), (1., -1., 0.)), None, (0., 90.)),
    }

    def testAngleBetweenVectorsFunction(self):
        for name, params in self.TESTS.items():
            refvector, vectors, norm, refangles = params
            with self.subTest(name):
                refangles = numpy.radians(refangles)

                refvector = numpy.array(refvector)
                vectors = numpy.array(vectors)
                if norm is not None:
                    norm = numpy.array(norm)

                testangles = utils.angleBetweenVectors(
                    refvector, vectors, norm)

                self.assertTrue(
                    numpy.allclose(testangles, refangles, atol=1e-5))


# Plane #######################################################################

class AssertNotificationContext(object):
    """Context that checks if an event.Notifier is sending events."""

    def __init__(self, notifier, count=1):
        """Initializer.

        :param event.Notifier notifier: The notifier to test.
        :param int count: The expected number of calls.
        """
        self._notifier = notifier
        self._callCount = None
        self._count = count

    def __enter__(self):
        self._callCount = 0
        self._notifier.addListener(self._callback)

    def __exit__(self, exc_type, exc_value, traceback):
        # Do not return True so exceptions are propagated
        self._notifier.removeListener(self._callback)
        assert self._callCount == self._count
        self._callCount = None

    def _callback(self, *args, **kwargs):
        self._callCount += 1


class TestPlaneParameters(ParametricTestCase):
    """Test Plane.parameters read/write and notifications."""

    PARAMETERS = {
        'unit normal': (1., 0., 0., 1.),
        'not unit normal': (1., 1., 0., 1.),
        'd = 0': (1., 0., 0., 0.)
    }

    def testParameters(self):
        """Check parameters read/write and notification."""
        plane = utils.Plane()

        for name, parameters in self.PARAMETERS.items():
            with self.subTest(name, parameters=parameters):
                with AssertNotificationContext(plane):
                    plane.parameters = parameters

                # Plane parameters are converted to have a unit normal
                normparams = parameters / numpy.linalg.norm(parameters[:3])
                self.assertTrue(numpy.allclose(plane.parameters, normparams))

    ZEROS_PARAMETERS = (
        (0., 0., 0., 0.),
        (0., 0., 0., 1.)
    )

    ZEROS = 0., 0., 0., 0.

    def testParametersNoPlane(self):
        """Test Plane.parameters with ||normal|| == 0 ."""
        plane = utils.Plane()
        plane.parameters = self.ZEROS

        for parameters in self.ZEROS_PARAMETERS:
            with self.subTest(parameters=parameters):
                with AssertNotificationContext(plane, count=0):
                    plane.parameters = parameters
                self.assertTrue(
                    numpy.allclose(plane.parameters, self.ZEROS, 0., 0.))


# unindexArrays ###############################################################

class TestUnindexArrays(ParametricTestCase):
    """Test unindexArrays function."""

    def testBasicModes(self):
        """Test for modes: points, lines and triangles"""
        indices = numpy.array((1, 2, 0))
        arrays = (numpy.array((0., 1., 2.)),
                  numpy.array(((0, 0), (1, 1), (2, 2))))
        refresults = (numpy.array((1., 2., 0.)),
                      numpy.array(((1, 1), (2, 2), (0, 0))))

        for mode in ('points', 'lines', 'triangles'):
            with self.subTest(mode=mode):
                testresults = utils.unindexArrays(mode, indices, *arrays)
                for ref, test in zip(refresults, testresults):
                    self.assertTrue(numpy.equal(ref, test).all())

    def testPackedLines(self):
        """Test for modes: line_strip, loop"""
        indices = numpy.array((1, 2, 0))
        arrays = (numpy.array((0., 1., 2.)),
                  numpy.array(((0, 0), (1, 1), (2, 2))))
        results = {
            'line_strip': (
                numpy.array((1., 2., 2., 0.)),
                numpy.array(((1, 1), (2, 2), (2, 2), (0, 0)))),
            'loop': (
                numpy.array((1., 2., 2., 0., 0., 1.)),
                numpy.array(((1, 1), (2, 2), (2, 2), (0, 0), (0, 0), (1, 1)))),
        }

        for mode, refresults in results.items():
            with self.subTest(mode=mode):
                testresults = utils.unindexArrays(mode, indices, *arrays)
                for ref, test in zip(refresults, testresults):
                    self.assertTrue(numpy.equal(ref, test).all())

    def testPackedTriangles(self):
        """Test for modes: triangle_strip, fan"""
        indices = numpy.array((1, 2, 0, 3))
        arrays = (numpy.array((0., 1., 2., 3.)),
                  numpy.array(((0, 0), (1, 1), (2, 2), (3, 3))))
        results = {
            'triangle_strip': (
                numpy.array((1., 2., 0., 2., 0., 3.)),
                numpy.array(((1, 1), (2, 2), (0, 0), (2, 2), (0, 0), (3, 3)))),
            'fan': (
                numpy.array((1., 2., 0., 1., 0., 3.)),
                numpy.array(((1, 1), (2, 2), (0, 0), (1, 1), (0, 0), (3, 3)))),
        }

        for mode, refresults in results.items():
            with self.subTest(mode=mode):
                testresults = utils.unindexArrays(mode, indices, *arrays)
                for ref, test in zip(refresults, testresults):
                    self.assertTrue(numpy.equal(ref, test).all())

    def testBadIndices(self):
        """Test with negative indices and indices higher than array length"""
        arrays = numpy.array((0, 1)), numpy.array((0, 1, 2))

        # negative indices
        with self.assertRaises(AssertionError):
            utils.unindexArrays('points', (-1, 0), *arrays)

        # Too high indices
        with self.assertRaises(AssertionError):
            utils.unindexArrays('points', (0, 10), *arrays)


# triangleNormals #############################################################

class TestTriangleNormals(ParametricTestCase):
    """Test triangleNormals function."""

    def test(self):
        """Test for modes: points, lines and triangles"""
        positions = numpy.array(
            ((0., 0., 0.), (1., 0., 0.), (0., 1., 0.),  # normal = Z
             (1., 1., 1.), (1., 2., 3.), (4., 5., 6.),  # Random triangle
             # Degenerated triangles:
             (0., 0., 0.), (1., 0., 0.), (2., 0., 0.),  # Colinear points
             (1., 1., 1.), (1., 1., 1.), (1., 1., 1.),  # All same point
             ),
            dtype='float32')

        normals = numpy.array(
            ((0., 0., 1.),
             (-0.40824829,  0.81649658, -0.40824829),
             (0., 0., 0.),
             (0., 0., 0.)),
            dtype='float32')

        testnormals = utils.trianglesNormal(positions)
        self.assertTrue(numpy.allclose(testnormals, normals))


# suite #######################################################################

def suite():
    testsuite = unittest.TestSuite()
    for test in (TestAngleBetweenVectors,
                 TestPlaneParameters,
                 TestUnindexArrays,
                 TestTriangleNormals):
        testsuite.addTest(
            unittest.defaultTestLoader.loadTestsFromTestCase(test))
    return testsuite


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2004-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Utility functions, toolbars and actions  to create profile on images
and stacks of images"""


__authors__ = ["V.A. Sole", "T. Vincent", "P. Knobel", "H. Payno"]
__license__ = "MIT"
__date__ = "12/04/2019"


import weakref

from .. import qt
from . import actions
from .tools.profile import core
from .tools.profile import manager
from .tools.profile import rois
from silx.gui.widgets.MultiModeAction import MultiModeAction

from silx.utils.deprecation import deprecated
from silx.utils.deprecation import deprecated_warning
from .tools import roi as roi_mdl
from silx.gui.plot import items


@deprecated(replacement="silx.gui.plot.tools.profile.createProfile", since_version="0.13.0")
def createProfile(roiInfo, currentData, origin, scale, lineWidth, method):
    return core.createProfile(roiInfo, currentData, origin,
                              scale, lineWidth, method)


class _CustomProfileManager(manager.ProfileManager):
    """This custom profile manager uses a single predefined profile window
    if it is specified. Else the behavior is the same as the default
    ProfileManager """

    def setProfileWindow(self, profileWindow):
        self.__profileWindow = profileWindow

    def createProfileWindow(self, plot, roi):
        if self.__profileWindow is not None:
            return self.__profileWindow
        else:
            return super(_CustomProfileManager, self).createProfileWindow(plot, roi)

    def clearProfileWindow(self, profileWindow):
        if self.__profileWindow is not None:
            self.__profileWindow.setProfile(None)
        else:
            return super(_CustomProfileManager, self).clearProfileWindow(profileWindow)


class ProfileToolBar(qt.QToolBar):
    """QToolBar providing profile tools operating on a :class:`PlotWindow`.

    Attributes:

    - plot: Associated :class:`PlotWindow` on which the profile line is drawn.
    - actionGroup: :class:`QActionGroup` of available actions.

    To run the following sample code, a QApplication must be initialized.
    First, create a PlotWindow and add a :class:`ProfileToolBar`.

    >>> from silx.gui.plot import PlotWindow
    >>> from silx.gui.plot.Profile import ProfileToolBar

    >>> plot = PlotWindow()  # Create a PlotWindow
    >>> toolBar = ProfileToolBar(plot=plot)  # Create a profile toolbar
    >>> plot.addToolBar(toolBar)  # Add it to plot
    >>> plot.show()  # To display the PlotWindow with the profile toolbar

    :param plot: :class:`PlotWindow` instance on which to operate.
    :param profileWindow: Plot widget instance where to
                          display the profile curve or None to create one.
    :param str title: See :class:`QToolBar`.
    :param parent: See :class:`QToolBar`.
    """

    def __init__(self, parent=None, plot=None, profileWindow=None,
                 title=None):
        super(ProfileToolBar, self).__init__(title, parent)
        assert plot is not None

        if title is not None:
            deprecated_warning("Attribute",
                               name="title",
                               reason="removed",
                               since_version="0.13.0",
                               only_once=True,
                               skip_backtrace_count=1)

        self._plotRef = weakref.ref(plot)

        # If a profileWindow is defined,
        # It will be used to display all the profiles
        self._manager = _CustomProfileManager(self, plot)
        self._manager.setProfileWindow(profileWindow)
        self._manager.setDefaultColorFromCursorColor(True)
        self._manager.setItemType(image=True)
        self._manager.setActiveItemTracking(True)

        # Actions
        self._browseAction = actions.mode.ZoomModeAction(plot, parent=self)
        self._browseAction.setVisible(False)
        self.freeLineAction = None
        self._createProfileActions()
        self._editor = self._manager.createEditorAction(self)

        # ActionGroup
        self.actionGroup = qt.QActionGroup(self)
        self.actionGroup.addAction(self._browseAction)
        self.actionGroup.addAction(self.hLineAction)
        self.actionGroup.addAction(self.vLineAction)
        self.actionGroup.addAction(self.lineAction)
        self.actionGroup.addAction(self._editor)

        modes = MultiModeAction(self)
        modes.addAction(self.hLineAction)
        modes.addAction(self.vLineAction)
        modes.addAction(self.lineAction)
        if self.freeLineAction is not None:
            modes.addAction(self.freeLineAction)
        modes.addAction(self.crossAction)
        self.__multiAction = modes

        # Add actions to ToolBar
        self.addAction(self._browseAction)
        self.addAction(modes)
        self.addAction(self._editor)
        self.addAction(self.clearAction)

        plot.sigActiveImageChanged.connect(self._activeImageChanged)
        self._activeImageChanged()

    def _createProfileActions(self):
        self.hLineAction = self._manager.createProfileAction(rois.ProfileImageHorizontalLineROI, self)
        self.vLineAction = self._manager.createProfileAction(rois.ProfileImageVerticalLineROI, self)
        self.lineAction = self._manager.createProfileAction(rois.ProfileImageLineROI, self)
        self.freeLineAction = self._manager.createProfileAction(rois.ProfileImageDirectedLineROI, self)
        self.crossAction = self._manager.createProfileAction(rois.ProfileImageCrossROI, self)
        self.clearAction = self._manager.createClearAction(self)

    def getPlotWidget(self):
        """The :class:`.PlotWidget` associated to the toolbar."""
        return self._plotRef()

    @property
    @deprecated(since_version="0.13.0", replacement="getPlotWidget()")
    def plot(self):
        return self.getPlotWidget()

    def _setRoiActionEnabled(self, itemKind, enabled):
        for action in self.__multiAction.getMenu().actions():
            if not isinstance(action, roi_mdl.CreateRoiModeAction):
                continue
            roiClass = action.getRoiClass()
            if issubclass(itemKind, roiClass.ITEM_KIND):
                action.setEnabled(enabled)

    def _activeImageChanged(self, previous=None, legend=None):
        """Handle active image change to toggle actions"""
        if legend is None:
            self._setRoiActionEnabled(items.ImageStack, False)
            self._setRoiActionEnabled(items.ImageBase, False)
        else:
            plot = self.getPlotWidget()
            image = plot.getActiveImage()
            # Disable for empty image
            enabled = image.getData(copy=False).size > 0
            self._setRoiActionEnabled(type(image), enabled)

    @property
    @deprecated(since_version="0.6.0")
    def browseAction(self):
        return self._browseAction

    @property
    @deprecated(replacement="getProfilePlot", since_version="0.5.0")
    def profileWindow(self):
        return self.getProfilePlot()

    def getProfileManager(self):
        """Return the manager of the profiles.

        :rtype: ProfileManager
        """
        return self._manager

    @deprecated(since_version="0.13.0")
    def getProfilePlot(self):
        """Return plot widget in which the profile curve or the
        profile image is plotted.
        """
        window = self.getProfileMainWindow()
        if window is None:
            return None
        return window.getCurrentPlotWidget()

    @deprecated(replacement="getProfileManager().getCurrentRoi().getProfileWindow()", since_version="0.13.0")
    def getProfileMainWindow(self):
        """Return window containing the profile curve widget.

        This can return None if no profile was computed.
        """
        roi = self._manager.getCurrentRoi()
        if roi is None:
            return None
        return roi.getProfileWindow()

    @property
    @deprecated(since_version="0.13.0")
    def overlayColor(self):
        """This method does nothing anymore. But could be implemented if needed.

        It was used to set color to use for the ROI.

        If set to None (the default), the overlay color is adapted to the
        active image colormap and changes if the active image colormap changes.
        """
        pass

    @overlayColor.setter
    @deprecated(since_version="0.13.0")
    def overlayColor(self, color):
        """This method does nothing anymore. But could be implemented if needed.
        """
        pass

    def clearProfile(self):
        """Remove profile curve and profile area."""
        self._manager.clearProfile()

    @deprecated(since_version="0.13.0")
    def updateProfile(self):
        """This method does nothing anymore. But could be implemented if needed.

        It was used to update the displayed profile and profile ROI.

        This uses the current active image of the plot and the current ROI.
        """
        pass

    @deprecated(replacement="clearProfile()", since_version="0.13.0")
    def hideProfileWindow(self):
        """Hide profile window.
        """
        self.clearProfile()

    @deprecated(since_version="0.13.0")
    def setProfileMethod(self, method):
        assert method in ('sum', 'mean')
        roi = self._manager.getCurrentRoi()
        if roi is None:
            raise RuntimeError("No profile ROI selected")
        roi.setProfileMethod(method)

    @deprecated(since_version="0.13.0")
    def getProfileMethod(self):
        roi = self._manager.getCurrentRoi()
        if roi is None:
            raise RuntimeError("No profile ROI selected")
        return roi.getProfileMethod()

    @deprecated(since_version="0.13.0")
    def getProfileOptionToolAction(self):
        return self._editor


class Profile3DToolBar(ProfileToolBar):
    def __init__(self, parent=None, stackview=None,
                 title=None):
        """QToolBar providing profile tools for an image or a stack of images.

        :param parent: the parent QWidget
        :param stackview: :class:`StackView` instance on which to operate.
        :param str title: See :class:`QToolBar`.
        :param parent: See :class:`QToolBar`.
        """
        # TODO: add param profileWindow (specify the plot used for profiles)
        super(Profile3DToolBar, self).__init__(parent=parent,
                                               plot=stackview.getPlotWidget())

        if title is not None:
            deprecated_warning("Attribute",
                               name="title",
                               reason="removed",
                               since_version="0.13.0",
                               only_once=True,
                               skip_backtrace_count=1)

        self.stackView = stackview
        """:class:`StackView` instance"""

    def _createProfileActions(self):
        self.hLineAction = self._manager.createProfileAction(rois.ProfileImageStackHorizontalLineROI, self)
        self.vLineAction = self._manager.createProfileAction(rois.ProfileImageStackVerticalLineROI, self)
        self.lineAction = self._manager.createProfileAction(rois.ProfileImageStackLineROI, self)
        self.crossAction = self._manager.createProfileAction(rois.ProfileImageStackCrossROI, self)
        self.clearAction = self._manager.createClearAction(self)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""QWidget displaying a 3D volume as a stack of 2D images.

The :class:`StackView` class implements this widget.

Basic usage of :class:`StackView` is through the following methods:

- :meth:`StackView.getColormap`, :meth:`StackView.setColormap` to update the
  default colormap to use and update the currently displayed image.
- :meth:`StackView.setStack` to update the displayed image.

The :class:`StackView` uses :class:`PlotWindow` and also
exposes a subset of the :class:`silx.gui.plot.Plot` API for further control
(plot title, axes labels, ...).

The :class:`StackViewMainWindow` class implements a widget that adds a status
bar displaying the 3D index and the value under the mouse cursor.

Example::

    import numpy
    import sys
    from silx.gui import qt
    from silx.gui.plot.StackView import StackViewMainWindow


    app = qt.QApplication(sys.argv[1:])

    # synthetic data, stack of 100 images of size 200x300
    mystack = numpy.fromfunction(
        lambda i, j, k: numpy.sin(i/15.) + numpy.cos(j/4.) + 2 * numpy.sin(k/6.),
        (100, 200, 300)
    )


    sv = StackViewMainWindow()
    sv.setColormap("jet", autoscale=True)
    sv.setStack(mystack)
    sv.setLabels(["1st dim (0-99)", "2nd dim (0-199)",
                  "3rd dim (0-299)"])
    sv.show()

    app.exec_()

"""

__authors__ = ["P. Knobel", "H. Payno"]
__license__ = "MIT"
__date__ = "10/10/2018"

import numpy
import logging

import silx
from silx.gui import qt
from .. import icons
from . import items, PlotWindow, actions
from .items.image import ImageStack
from ..colors import Colormap
from ..colors import cursorColorForColormap
from .tools import LimitsToolBar
from .Profile import Profile3DToolBar
from ..widgets.FrameBrowser import HorizontalSliderWithBrowser

from silx.gui.plot.actions import control as actions_control
from silx.gui.plot.actions import io as silx_io
from silx.io.nxdata import save_NXdata
from silx.utils.array_like import DatasetView, ListOfImages
from silx.math import calibration
from silx.utils.deprecation import deprecated_warning
from silx.utils.deprecation import deprecated

import h5py
from silx.io.utils import is_dataset

_logger = logging.getLogger(__name__)


class StackView(qt.QMainWindow):
    """Stack view widget, to display and browse through stack of
    images.

    The profile tool can be switched to "3D" mode, to compute the profile
    on each image of the stack (not only the active image currently displayed)
    and display the result as a slice.

    :param QWidget parent: the Qt parent, or None
    :param backend: The backend to use for the plot (default: matplotlib).
                    See :class:`.PlotWidget` for the list of supported backend.
    :type backend: str or :class:`BackendBase.BackendBase`
    :param bool resetzoom: Toggle visibility of reset zoom action.
    :param bool autoScale: Toggle visibility of axes autoscale actions.
    :param bool logScale: Toggle visibility of axes log scale actions.
    :param bool grid: Toggle visibility of grid mode action.
    :param bool colormap: Toggle visibility of colormap action.
    :param bool aspectRatio: Toggle visibility of aspect ratio button.
    :param bool yInverted: Toggle visibility of Y axis direction button.
    :param bool copy: Toggle visibility of copy action.
    :param bool save: Toggle visibility of save action.
    :param bool print_: Toggle visibility of print action.
    :param bool control: True to display an Options button with a sub-menu
                         to show legends, toggle crosshair and pan with arrows.
                         (Default: False)
    :param position: True to display widget with (x, y) mouse position
                     (Default: False).
                     It also supports a list of (name, funct(x, y)->value)
                     to customize the displayed values.
                     See :class:`silx.gui.plot.PlotTools.PositionInfo`.
    :param bool mask: Toggle visibilty of mask action.
    """
    # Qt signals
    valueChanged = qt.Signal(object, object, object)
    """Signals that the data value under the cursor has changed.

    It provides: row, column, data value.
    """

    sigPlaneSelectionChanged = qt.Signal(int)
    """Signal emitted when there is a change is perspective/displayed axes.

    It provides the perspective as an integer, with the following meaning:

        - 0: axis Y is the 2nd dimension, axis X is the 3rd dimension
        - 1: axis Y is the 1st dimension, axis X is the 3rd dimension
        - 2: axis Y is the 1st dimension, axis X is the 2nd dimension
    """

    sigStackChanged = qt.Signal(int)
    """Signal emitted when the stack is changed.
    This happens when a new volume is loaded, or when the current volume
    is transposed (change in perspective).

    The signal provides the size (number of pixels) of the stack.
    This will be 0 if the stack is cleared, else it will be a positive
    integer.
    """

    sigFrameChanged = qt.Signal(int)
    """Signal emitter when the frame number has changed.

    This signal provides the current frame number.
    """

    IMAGE_STACK_FILTER_NXDATA = 'Stack of images as NXdata (%s)' % silx_io._NEXUS_HDF5_EXT_STR


    def __init__(self, parent=None, resetzoom=True, backend=None,
                 autoScale=False, logScale=False, grid=False,
                 colormap=True, aspectRatio=True, yinverted=True,
                 copy=True, save=True, print_=True, control=False,
                 position=None, mask=True):
        qt.QMainWindow.__init__(self, parent)
        if parent is not None:
            # behave as a widget
            self.setWindowFlags(qt.Qt.Widget)
        else:
            self.setWindowTitle('StackView')

        self._stack = None
        """Loaded stack, as a 3D array, a 3D dataset or a list of 2D arrays."""
        self.__transposed_view = None
        """View on :attr:`_stack` with the axes sorted, to have
        the orthogonal dimension first"""
        self._perspective = 0
        """Orthogonal dimension (depth) in :attr:`_stack`"""

        self._stackItem = ImageStack()
        """Hold the item displaying the stack"""
        imageLegend = '__StackView__image' + str(id(self))
        self._stackItem.setName(imageLegend)

        self.__autoscaleCmap = False
        """Flag to disable/enable colormap auto-scaling
        based on the min/max values of the entire 3D volume"""
        self.__dimensionsLabels = ["Dimension 0", "Dimension 1",
                                   "Dimension 2"]
        """These labels are displayed on the X and Y axes.
        :meth:`setLabels` updates this attribute."""

        self._first_stack_dimension = 0
        """Used for dimension labels and combobox"""

        self._titleCallback = self._defaultTitleCallback
        """Function returning the plot title based on the frame index.
        It can be set to a custom function using :meth:`setTitleCallback`"""

        self.calibrations3D = (calibration.NoCalibration(),
                               calibration.NoCalibration(),
                               calibration.NoCalibration())

        central_widget = qt.QWidget(self)

        self._plot = PlotWindow(parent=central_widget, backend=backend,
                                resetzoom=resetzoom, autoScale=autoScale,
                                logScale=logScale, grid=grid,
                                curveStyle=False, colormap=colormap,
                                aspectRatio=aspectRatio, yInverted=yinverted,
                                copy=copy, save=save, print_=print_,
                                control=control, position=position,
                                roi=False, mask=mask)
        self._plot.addItem(self._stackItem)
        self._plot.getIntensityHistogramAction().setVisible(True)
        self.sigInteractiveModeChanged = self._plot.sigInteractiveModeChanged
        self.sigActiveImageChanged = self._plot.sigActiveImageChanged
        self.sigPlotSignal = self._plot.sigPlotSignal

        if silx.config.DEFAULT_PLOT_IMAGE_Y_AXIS_ORIENTATION == 'downward':
            self._plot.getYAxis().setInverted(True)

        self._addColorBarAction()

        self._profileToolBar = Profile3DToolBar(parent=self._plot,
                                                stackview=self)
        self._plot.addToolBar(self._profileToolBar)
        self._plot.getXAxis().setLabel('Columns')
        self._plot.getYAxis().setLabel('Rows')
        self._plot.sigPlotSignal.connect(self._plotCallback)
        self._plot.getSaveAction().setFileFilter('image', self.IMAGE_STACK_FILTER_NXDATA, func=self._saveImageStack, appendToFile=True)

        self.__planeSelection = PlanesWidget(self._plot)
        self.__planeSelection.sigPlaneSelectionChanged.connect(self.setPerspective)

        self._browser_label = qt.QLabel("Image index (Dim0):")

        self._browser = HorizontalSliderWithBrowser(central_widget)
        self._browser.setRange(0, 0)
        self._browser.valueChanged[int].connect(self.__updateFrameNumber)
        self._browser.setEnabled(False)

        layout = qt.QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._plot, 0, 0, 1, 3)
        layout.addWidget(self.__planeSelection, 1, 0)
        layout.addWidget(self._browser_label, 1, 1)
        layout.addWidget(self._browser, 1, 2)

        central_widget.setLayout(layout)
        self.setCentralWidget(central_widget)

        # clear profile lines when the perspective changes (plane browsed changed)
        self.__planeSelection.sigPlaneSelectionChanged.connect(
            self._profileToolBar.clearProfile)

    def _saveImageStack(self, plot, filename, nameFilter):
        """Save all images from the stack into a volume.

        :param str filename: The name of the file to write
        :param str nameFilter: The selected name filter
        :return: False if format is not supported or save failed,
                 True otherwise.
        :raises: ValueError if nameFilter is invalid
        """
        if not nameFilter == self.IMAGE_STACK_FILTER_NXDATA:
            raise ValueError('Wrong callback')
        entryPath = silx_io.SaveAction._selectWriteableOutputGroup(filename, parent=self)
        if entryPath is None:
            return False
        return save_NXdata(filename,
                           nxentry_name=entryPath,
                           signal=self.getStack(copy=False, returnNumpyArray=True)[0],
                           signal_name="image_stack")

    def _addColorBarAction(self):
        self._plot.getColorBarWidget().setVisible(True)
        actions = self._plot.toolBar().actions()
        for index, action in enumerate(actions):
            if action is self._plot.getColormapAction():
                break
        self._colorbarAction = actions_control.ColorBarAction(self._plot, self._plot)
        self._plot.toolBar().insertAction(actions[index + 1], self._colorbarAction)

    def _plotCallback(self, eventDict):
        """Callback for plot events.

        Emit :attr:`valueChanged` signal, with (x, y, value) tuple of the
        cursor location in the plot."""
        if eventDict['event'] == 'mouseMoved':
            activeImage = self.getActiveImage()
            if activeImage is not None:
                data = activeImage.getData()
                height, width = data.shape

                # Get corresponding coordinate in image
                origin = activeImage.getOrigin()
                scale = activeImage.getScale()
                x = int((eventDict['x'] - origin[0]) / scale[0])
                y = int((eventDict['y'] - origin[1]) / scale[1])

                if 0 <= x < width and 0 <= y < height:
                    self.valueChanged.emit(float(x), float(y),
                                           data[y][x])
                else:
                    self.valueChanged.emit(float(x), float(y),
                                           None)

    def getPerspective(self):
        """Returns the index of the dimension the stack is browsed with

        Possible values are: 0, 1, or 2.

        :rtype: int
        """
        return self._perspective

    def setPerspective(self, perspective):
        """Set the index of the dimension the stack is browsed with:

        - slice plane Dim1-Dim2: perspective 0
        - slice plane Dim0-Dim2: perspective 1
        - slice plane Dim0-Dim1: perspective 2

        :param int perspective: Orthogonal dimension number (0, 1, or 2)
        """
        if perspective == self._perspective:
            return
        else:
            if perspective > 2 or perspective < 0:
                raise ValueError(
                    "Perspective must be 0, 1 or 2, not %s" % perspective)

            self._perspective = int(perspective)
            self.__createTransposedView()
            self.__updateFrameNumber(self._browser.value())
            self._plot.resetZoom()
            self.__updatePlotLabels()
            self._updateTitle()
            self._browser_label.setText("Image index (Dim%d):" %
                                        (self._first_stack_dimension + perspective))

            self.sigPlaneSelectionChanged.emit(perspective)
            self.sigStackChanged.emit(self._stack.size if
                                      self._stack is not None else 0)
            self.__planeSelection.sigPlaneSelectionChanged.disconnect(self.setPerspective)
            self.__planeSelection.setPerspective(self._perspective)
            self.__planeSelection.sigPlaneSelectionChanged.connect(self.setPerspective)

    def __updatePlotLabels(self):
        """Update plot axes labels depending on perspective"""
        y, x = (1, 2) if self._perspective == 0 else \
            (0, 2) if self._perspective == 1 else (0, 1)
        self.setGraphXLabel(self.__dimensionsLabels[x])
        self.setGraphYLabel(self.__dimensionsLabels[y])

    def __createTransposedView(self):
        """Create the new view on the stack depending on the perspective
        (set orthogonal axis browsed on the viewer as first dimension)
        """
        assert self._stack is not None
        assert 0 <= self._perspective < 3

        # ensure we have the stack encapsulated in an array-like object
        # having a transpose() method
        if isinstance(self._stack, numpy.ndarray):
            self.__transposed_view = self._stack

        elif is_dataset(self._stack) or isinstance(self._stack, DatasetView):
            self.__transposed_view = DatasetView(self._stack)

        elif isinstance(self._stack, ListOfImages):
            self.__transposed_view = ListOfImages(self._stack)

        # transpose the array-like object if necessary
        if self._perspective == 1:
            self.__transposed_view = self.__transposed_view.transpose((1, 0, 2))
        elif self._perspective == 2:
            self.__transposed_view = self.__transposed_view.transpose((2, 0, 1))

        self._browser.setRange(0, self.__transposed_view.shape[0] - 1)
        self._browser.setValue(0)

        # Update the item structure
        self._stackItem.setStackData(self.__transposed_view, 0, copy=False)
        self._stackItem.setColormap(self.getColormap())
        self._stackItem.setOrigin(self._getImageOrigin())
        self._stackItem.setScale(self._getImageScale())

    def __updateFrameNumber(self, index):
        """Update the current image.

        :param index: index of the frame to be displayed
        """
        if self.__transposed_view is None:
            # no data set
            return

        self._stackItem.setStackPosition(index)

        self._updateTitle()
        self.sigFrameChanged.emit(index)

    def _set3DScaleAndOrigin(self, calibrations):
        """Set scale and origin for all 3 axes, to be used when plotting
        an image.

        See setStack for parameter documentation
        """
        if calibrations is None:
            self.calibrations3D = (calibration.NoCalibration(),
                                   calibration.NoCalibration(),
                                   calibration.NoCalibration())
        else:
            self.calibrations3D = []
            for i, calib in enumerate(calibrations):
                if hasattr(calib, "__len__") and len(calib) == 2:
                    calib = calibration.LinearCalibration(calib[0], calib[1])
                elif calib is None:
                    calib = calibration.NoCalibration()
                elif not isinstance(calib, calibration.AbstractCalibration):
                    raise TypeError("calibration must be a 2-tuple, None or" +
                                    " an instance of an AbstractCalibration " +
                                    "subclass")
                elif not calib.is_affine():
                    _logger.warning(
                            "Calibration for dimension %d is not linear, "
                            "it will be ignored for scaling the graph axes.",
                            i)
                self.calibrations3D.append(calib)

    def getCalibrations(self, order='array'):
        """Returns currently used calibrations for each axis

        Returned calibrations might differ from the ones that were set as
        non-linear calibrations used for image axes are temporarily ignored.

        :param str order:
            'array' to sort calibrations as data array (dim0, dim1, dim2),
            'axes' to sort calibrations as currently selected x, y and z axes.
        :return: Calibrations ordered depending on order
        :rtype: List[~silx.math.calibration.AbstractCalibration]
        """
        assert order in ('array', 'axes')
        calibs = []

        # filter out non-linear calibration for graph axes
        for index, calib in enumerate(self.calibrations3D):
            if index != self._perspective and not calib.is_affine():
                calib = calibration.NoCalibration()
            calibs.append(calib)

        if order == 'axes':  # Move 'z' axis to the end
            xy_dims = [d for d in (0, 1, 2) if d != self._perspective]
            calibs = [calibs[max(xy_dims)],
                      calibs[min(xy_dims)],
                      calibs[self._perspective]]

        return tuple(calibs)

    def _getImageScale(self):
        """
        :return: 2-tuple (XScale, YScale) for current image view
        """
        xcalib, ycalib, _zcalib = self.getCalibrations(order='axes')
        return xcalib.get_slope(), ycalib.get_slope()

    def _getImageOrigin(self):
        """
        :return: 2-tuple (XOrigin, YOrigin) for current image view
        """
        xcalib, ycalib, _zcalib = self.getCalibrations(order='axes')
        return xcalib(0), ycalib(0)

    def _getImageZ(self, index):
        """
        :param idx: 0-based image index in the stack
        :return: calibrated Z value corresponding to the image idx
        """
        _xcalib, _ycalib, zcalib = self.getCalibrations(order='axes')
        return zcalib(index)

    def _updateTitle(self):
        frame_idx = self._browser.value()
        self._plot.setGraphTitle(self._titleCallback(frame_idx))

    def _defaultTitleCallback(self, index):
        return "Image z=%g" % self._getImageZ(index)

    # public API, stack specific methods
    def setStack(self, stack, perspective=None, reset=True, calibrations=None):
        """Set the 3D stack.

        The perspective parameter is used to define which dimension of the 3D
        array is to be used as frame index. The lowest remaining dimension
        number is the row index of the displayed image (Y axis), and the highest
        remaining dimension is the column index (X axis).

        :param stack: 3D stack, or `None` to clear plot.
        :type stack: 3D numpy.ndarray, or 3D h5py.Dataset, or list/tuple of 2D
            numpy arrays, or None.
        :param int perspective: Dimension for the frame index: 0, 1 or 2.
            Use ``None`` to keep the current perspective (default).
        :param bool reset: Whether to reset zoom or not.
        :param calibrations: Sequence of 3 calibration objects for each axis.
            These objects can be a subclass of :class:`AbstractCalibration`,
            or 2-tuples *(a, b)* where *a* is the y-intercept and *b* is the
            slope of a linear calibration (:math:`x \\mapsto a + b x`)
        """
        if stack is None:
            self.clear()
            self.sigStackChanged.emit(0)
            return

        self._set3DScaleAndOrigin(calibrations)

        # stack as list of 2D arrays: must be converted into an array_like
        if not isinstance(stack, numpy.ndarray):
            if not is_dataset(stack):
                try:
                    assert hasattr(stack, "__len__")
                    for img in stack:
                        assert hasattr(img, "shape")
                        assert len(img.shape) == 2
                except AssertionError:
                    raise ValueError(
                        "Stack must be a 3D array/dataset or a list of " +
                        "2D arrays.")
                stack = ListOfImages(stack)

        assert len(stack.shape) == 3, "data must be 3D"

        self._stack = stack
        self.__createTransposedView()

        perspective_changed = False
        if perspective not in [None, self._perspective]:
            perspective_changed = True
            self.setPerspective(perspective)

        if self.__autoscaleCmap:
            self.scaleColormapRangeToStack()

        # init plot
        self._stackItem.setStackData(self.__transposed_view, 0, copy=False)
        self._stackItem.setColormap(self.getColormap())
        self._stackItem.setOrigin(self._getImageOrigin())
        self._stackItem.setScale(self._getImageScale())
        self._stackItem.setVisible(True)

        # Put back the item in the plot in case it was cleared
        exists = self._plot.getImage(self._stackItem.getName())
        if exists is None:
            self._plot.addItem(self._stackItem)

        self._plot.setActiveImage(self._stackItem.getName())
        self.__updatePlotLabels()
        self._updateTitle()

        if reset:
            self._plot.resetZoom()

        # enable and init browser
        self._browser.setEnabled(True)

        if not perspective_changed:    # avoid double signal (see self.setPerspective)
            self.sigStackChanged.emit(stack.size)

    def getStack(self, copy=True, returnNumpyArray=False):
        """Get the original stack, as a 3D array or dataset.

        The output has the form: [data, params]
        where params is a dictionary containing display parameters.

        :param bool copy: If True (default), then the object is copied
            and returned as a numpy array.
            Else, a reference to original data is returned, if possible.
            If the original data is not a numpy array and parameter
            returnNumpyArray is True, a copy will be made anyway.
        :param bool returnNumpyArray: If True, the returned object is
            guaranteed to be a numpy array.
        :return: 3D stack and parameters.
        :rtype: (numpy.ndarray, dict)
        """
        if self._stack is None:
            return None

        image = self._stackItem
        colormap = image.getColormap()

        params = {
            'info': image.getInfo(),
            'origin': image.getOrigin(),
            'scale': image.getScale(),
            'z': image.getZValue(),
            'selectable': image.isSelectable(),
            'draggable': image.isDraggable(),
            'colormap': colormap,
            'xlabel': image.getXLabel(),
            'ylabel': image.getYLabel(),
        }
        if returnNumpyArray or copy:
            return numpy.array(self._stack, copy=copy), params

        # if a list of 2D arrays was cast into a ListOfImages,
        # return the original list
        if isinstance(self._stack, ListOfImages):
            return self._stack.images, params

        return self._stack, params

    def getCurrentView(self, copy=True, returnNumpyArray=False):
        """Get the stack, as it is currently displayed.

        The first index of the returned stack is always the frame
        index. If the perspective has been changed in the widget since the
        data was first loaded, this will be reflected in the order of the
        dimensions of the returned object.

        The output has the form: [data, params]
        where params is a dictionary containing display parameters.

        :param bool copy: If True (default), then the object is copied
            and returned as a numpy array.
            Else, a reference to original data is returned, if possible.
            If the original data is not a numpy array and parameter
            `returnNumpyArray` is `True`, a copy will be made anyway.
        :param bool returnNumpyArray: If `True`, the returned object is
            guaranteed to be a numpy array.
        :return: 3D stack and parameters.
        :rtype: (numpy.ndarray, dict)
        """
        image = self.getActiveImage()
        if image is None:
            return None

        if isinstance(image, items.ColormapMixIn):
            colormap = image.getColormap()
        else:
            colormap = None

        params = {
            'info': image.getInfo(),
            'origin': image.getOrigin(),
            'scale': image.getScale(),
            'z': image.getZValue(),
            'selectable': image.isSelectable(),
            'draggable': image.isDraggable(),
            'colormap': colormap,
            'xlabel': image.getXLabel(),
            'ylabel': image.getYLabel(),
        }
        if returnNumpyArray or copy:
            return numpy.array(self.__transposed_view, copy=copy), params
        return self.__transposed_view, params

    def setFrameNumber(self, number):
        """Set the frame selection to a specific value

        :param int number: Number of the frame
        """
        self._browser.setValue(number)

    def getFrameNumber(self):
        """Set the frame selection to a specific value

        :return: Index of currently displayed frame
        :rtype: int
        """
        return self._browser.value()

    def setFirstStackDimension(self, first_stack_dimension):
        """When viewing the last 3 dimensions of an n-D array (n>3), you can
        use this method to change the text in the combobox.

        For instance, for a 7-D array, first stack dim is 4, so the default
        "Dim1-Dim2" text should be replaced with "Dim5-Dim6" (dimensions
        numbers are 0-based).

        :param int first_stack_dim: First stack dimension (n-3) when viewing the
            last 3 dimensions of an n-D array.
        """
        old_state = self.__planeSelection.blockSignals(True)
        self.__planeSelection.setFirstStackDimension(first_stack_dimension)
        self.__planeSelection.blockSignals(old_state)
        self._first_stack_dimension = first_stack_dimension
        self._browser_label.setText("Image index (Dim%d):" % first_stack_dimension)

    def setTitleCallback(self, callback):
        """Set a user defined function to generate the plot title based on the
        image/frame index.

        The callback function must accept an integer as a its first positional
        parameter and must not require any other mandatory parameter.
        It must return a string.

        To switch back the default behavior, you can pass ``None``::

            mystackview.setTitleCallback(None)

        To have no title, pass a function that returns an empty string::

            mystackview.setTitleCallback(lambda idx: "")

        :param callback: Callback function generating the stack title based
            on the frame number.
        """

        if callback is None:
            self._titleCallback = self._defaultTitleCallback
        elif callable(callback):
            self._titleCallback = callback
        else:
            raise TypeError("Provided callback is not callable")
        self._updateTitle()

    def clear(self):
        """Clear the widget:

         - clear the plot
         - clear the loaded data volume
        """
        self._stack = None
        self.__transposed_view = None
        self._perspective = 0
        self._browser.setEnabled(False)
        # reset browser range
        self._browser.setRange(0, 0)
        self._plot.clear()

    def setLabels(self, labels=None):
        """Set the labels to be displayed on the plot axes.

        You must provide a sequence of 3 strings, corresponding to the 3
        dimensions of the original data volume.
        The proper label will automatically be selected for each plot axis
        when the volume is rotated (when different axes are selected as the
        X and Y axes).

        :param List[str] labels: 3 labels corresponding to the 3 dimensions
             of the data volumes.
        """

        default_labels = ["Dimension %d" % self._first_stack_dimension,
                          "Dimension %d" % (self._first_stack_dimension + 1),
                          "Dimension %d" % (self._first_stack_dimension + 2)]
        if labels is None:
            new_labels = default_labels
        else:
            # filter-out None
            new_labels = []
            for i, label in enumerate(labels):
                new_labels.append(label or default_labels[i])

        self.__dimensionsLabels = new_labels
        self.__updatePlotLabels()

    def getLabels(self):
        """Return dimension labels displayed on the plot axes

        :return: List of three strings corresponding to the 3 dimensions
            of the stack: (name_dim0, name_dim1, name_dim2)
        """
        return self.__dimensionsLabels

    def getColormap(self):
        """Get the current colormap description.

        :return: A description of the current colormap.
                 See :meth:`setColormap` for details.
        :rtype: dict
        """
        # "default" colormap used by addImage when image is added without
        # specifying a special colormap
        return self._plot.getDefaultColormap()

    def scaleColormapRangeToStack(self):
        """Scale colormap range according to current stack data.

        If no stack has been set through :meth:`setStack`, this has no effect.

        The range scaling mode is given by current :class:`Colormap`'s
        :meth:`Colormap.getAutoscaleMode`.
        """
        stack = self.getStack(copy=False, returnNumpyArray=True)
        if stack is None:
            return  # No-op

        colormap = self.getColormap()
        vmin, vmax = colormap.getColormapRange(data=stack[0])
        colormap.setVRange(vmin=vmin, vmax=vmax)

    def setColormap(self, colormap=None, normalization=None,
                    autoscale=None, vmin=None, vmax=None, colors=None):
        """Set the colormap and update active image.

        Parameters that are not provided are taken from the current colormap.

        The colormap parameter can also be a dict with the following keys:

        - *name*: string. The colormap to use:
          'gray', 'reversed gray', 'temperature', 'red', 'green', 'blue'.
        - *normalization*: string. The mapping to use for the colormap:
          either 'linear' or 'log'.
        - *autoscale*: bool. Whether to use autoscale (True) or range
          provided by keys
          'vmin' and 'vmax' (False).
        - *vmin*: float. The minimum value of the range to use if 'autoscale'
          is False.
        - *vmax*: float. The maximum value of the range to use if 'autoscale'
          is False.
        - *colors*: optional. Nx3 or Nx4 array of float in [0, 1] or uint8.
                    List of RGB or RGBA colors to use (only if name is None)

        :param colormap: Name of the colormap in
            'gray', 'reversed gray', 'temperature', 'red', 'green', 'blue'.
            Or a :class`.Colormap` object.
        :type colormap: dict or str.
        :param str normalization: Colormap mapping: 'linear' or 'log'.
        :param bool autoscale: Whether to use autoscale or [vmin, vmax] range.
            Default value of autoscale is False. This option is not compatible
            with h5py datasets.
        :param float vmin: The minimum value of the range to use if
                           'autoscale' is False.
        :param float vmax: The maximum value of the range to use if
                           'autoscale' is False.
        :param numpy.ndarray colors: Only used if name is None.
            Custom colormap colors as Nx3 or Nx4 RGB or RGBA arrays
        """
        # if is a colormap object or a dictionary
        if isinstance(colormap, Colormap) or isinstance(colormap, dict):
            # Support colormap parameter as a dict
            errmsg = "If colormap is provided as a Colormap object, all other parameters"
            errmsg += " must not be specified when calling setColormap"
            assert normalization is None, errmsg
            assert autoscale is None, errmsg
            assert vmin is None, errmsg
            assert vmax is None, errmsg
            assert colors is None, errmsg

            if isinstance(colormap, dict):
                reason = 'colormap parameter should now be an object'
                replacement = 'Colormap()'
                since_version = '0.6'
                deprecated_warning(type_='function',
                                   name='setColormap',
                                   reason=reason,
                                   replacement=replacement,
                                   since_version=since_version)
                _colormap = Colormap._fromDict(colormap)
            else:
                _colormap = colormap
        else:
            norm = normalization if normalization is not None else 'linear'
            name = colormap if colormap is not None else 'gray'
            _colormap = Colormap(name=name,
                                 normalization=norm,
                                 vmin=vmin,
                                 vmax=vmax,
                                 colors=colors)

            if autoscale is not None:
                deprecated_warning(
                    type_='function',
                    name='setColormap',
                    reason='autoscale argument is replaced by a method',
                    replacement='scaleColormapRangeToStack',
                    since_version='0.14')
            self.__autoscaleCmap = bool(autoscale)

        cursorColor = cursorColorForColormap(_colormap.getName())
        self._plot.setInteractiveMode('zoom', color=cursorColor)

        self._plot.setDefaultColormap(_colormap)

        # Update active image colormap
        activeImage = self.getActiveImage()
        if isinstance(activeImage, items.ColormapMixIn):
            activeImage.setColormap(self.getColormap())

        if self.__autoscaleCmap:
            # scaleColormapRangeToStack needs to be called **after**
            # setDefaultColormap so getColormap returns the right colormap
            self.scaleColormapRangeToStack()


    @deprecated(replacement="getPlotWidget", since_version="0.13")
    def getPlot(self):
        return self.getPlotWidget()

    def getPlotWidget(self):
        """Return the :class:`PlotWidget`.

        This gives access to advanced plot configuration options.
        Be warned that modifying the plot can cause issues, and some changes
        you make to the plot could be overwritten by the :class:`StackView`
        widget's internal methods and callbacks.

        :return: instance of :class:`PlotWidget` used in widget
        """
        return self._plot

    def setOptionVisible(self, isVisible):
        """
        Set the visibility of the browsing options.

        :param bool isVisible: True to have the options visible, else False
        """
        self._browser.setVisible(isVisible)
        self.__planeSelection.setVisible(isVisible)

    # proxies to PlotWidget or PlotWindow methods
    def getProfileToolbar(self):
        """Profile tools attached to this plot
        """
        return self._profileToolBar

    def getGraphTitle(self):
        """Return the plot main title as a str.
        """
        return self._plot.getGraphTitle()

    def setGraphTitle(self, title=""):
        """Set the plot main title.

        :param str title: Main title of the plot (default: '')
        """
        return self._plot.setGraphTitle(title)

    def getGraphXLabel(self):
        """Return the current horizontal axis label as a str.
        """
        return self._plot.getXAxis().getLabel()

    def setGraphXLabel(self, label=None):
        """Set the plot horizontal axis label.

        :param str label: The horizontal axis label
        """
        if label is None:
            label = self.__dimensionsLabels[1 if self._perspective == 2 else 2]
        self._plot.getXAxis().setLabel(label)

    def getGraphYLabel(self, axis='left'):
        """Return the current vertical axis label as a str.

        :param str axis: The Y axis for which to get the label (left or right)
        """
        return self._plot.getYAxis().getLabel(axis)

    def setGraphYLabel(self, label=None, axis='left'):
        """Set the vertical axis label on the plot.

        :param str label: The Y axis label
        :param str axis: The Y axis for which to set the label (left or right)
        """
        if label is None:
            label = self.__dimensionsLabels[1 if self._perspective == 0 else 0]
        self._plot.getYAxis(axis=axis).setLabel(label)

    def resetZoom(self):
        """Reset the plot limits to the bounds of the data and redraw the plot.

        This method is a simple proxy to the legacy :class:`PlotWidget` method
        of the same name. Using the object oriented approach is now
        preferred::

            stackview.getPlot().resetZoom()
        """
        self._plot.resetZoom()

    def setYAxisInverted(self, flag=True):
        """Set the Y axis orientation.

        This method is a simple proxy to the legacy :class:`PlotWidget` method
        of the same name. Using the object oriented approach is now
        preferred::

            stackview.getPlot().setYAxisInverted(flag)

        :param bool flag: True for Y axis going from top to bottom,
                          False for Y axis going from bottom to top
        """
        self._plot.setYAxisInverted(flag)

    def isYAxisInverted(self):
        """Return True if Y axis goes from top to bottom, False otherwise.

        This method is a simple proxy to the legacy :class:`PlotWidget` method
        of the same name. Using the object oriented approach is now
        preferred::

            stackview.getPlot().isYAxisInverted()"""
        return self._plot.isYAxisInverted()

    def getSupportedColormaps(self):
        """Get the supported colormap names as a tuple of str.

        The list should at least contain and start by:
        ('gray', 'reversed gray', 'temperature', 'red', 'green', 'blue')

        This method is a simple proxy to the legacy :class:`PlotWidget` method
        of the same name. Using the object oriented approach is now
        preferred::

            stackview.getPlot().getSupportedColormaps()
        """
        return self._plot.getSupportedColormaps()

    def isKeepDataAspectRatio(self):
        """Returns whether the plot is keeping data aspect ratio or not.

        This method is a simple proxy to the legacy :class:`PlotWidget` method
        of the same name. Using the object oriented approach is now
        preferred::

            stackview.getPlot().isKeepDataAspectRatio()"""
        return self._plot.isKeepDataAspectRatio()

    def setKeepDataAspectRatio(self, flag=True):
        """Set whether the plot keeps data aspect ratio or not.

        This method is a simple proxy to the legacy :class:`PlotWidget` method
        of the same name. Using the object oriented approach is now
        preferred::

            stackview.getPlot().setKeepDataAspectRatio(flag)

        :param bool flag: True to respect data aspect ratio
        """
        self._plot.setKeepDataAspectRatio(flag)

    # kind of private methods, but needed by Profile
    def getActiveImage(self, just_legend=False):
        """Returns the stack image object.
        """
        if just_legend:
            return self._stackItem.getName()
        return self._stackItem

    def getColorBarAction(self):
        """Returns the action managing the visibility of the colorbar.

        .. warning:: to show/hide the plot colorbar call directly the ColorBar
            widget using getColorBarWidget()

        :rtype: QAction
        """
        return self._colorbarAction

    def remove(self, legend=None,
               kind=('curve', 'image', 'item', 'marker')):
        """See :meth:`Plot.Plot.remove`"""
        self._plot.remove(legend, kind)

    def setInteractiveMode(self, *args, **kwargs):
        """
        See :meth:`Plot.Plot.setInteractiveMode`
        """
        self._plot.setInteractiveMode(*args, **kwargs)

    @deprecated(replacement="addShape", since_version="0.13")
    def addItem(self, *args, **kwargs):
        self.addShape(*args, **kwargs)

    def addShape(self, *args, **kwargs):
        """
        See :meth:`Plot.Plot.addShape`
        """
        self._plot.addShape(*args, **kwargs)


class PlanesWidget(qt.QWidget):
    """Widget for the plane/perspective selection

    :param parent: the parent QWidget
    """
    sigPlaneSelectionChanged = qt.Signal(int)

    def __init__(self, parent):
        super(PlanesWidget, self).__init__(parent)

        self.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum)
        layout0 = qt.QHBoxLayout()
        self.setLayout(layout0)
        layout0.setContentsMargins(0, 0, 0, 0)

        layout0.addWidget(qt.QLabel("Axes selection:"))

        # By default, the first dimension (dim0) is the frame index/depth/z,
        # the second dimension is the image row number/y axis
        # and the third dimension is the image column index/x axis

        # 1
        # | 0
        # |/__2
        self.qcbAxisSelection = qt.QComboBox(self)
        self._setCBChoices(first_stack_dimension=0)
        self.qcbAxisSelection.currentIndexChanged[int].connect(
            self.__planeSelectionChanged)

        layout0.addWidget(self.qcbAxisSelection)

    def __planeSelectionChanged(self, idx):
        """Callback function when the combobox selection changes

        idx is the dimension number orthogonal to the slice plane,
        following the convention:

          - slice plane Dim1-Dim2: perspective 0
          - slice plane Dim0-Dim2: perspective 1
          - slice plane Dim0-Dim1: perspective 2
        """
        self.sigPlaneSelectionChanged.emit(idx)

    def _setCBChoices(self, first_stack_dimension):
        self.qcbAxisSelection.clear()

        dim1dim2 = 'Dim%d-Dim%d' % (first_stack_dimension + 1,
                                    first_stack_dimension + 2)
        dim0dim2 = 'Dim%d-Dim%d' % (first_stack_dimension,
                                    first_stack_dimension + 2)
        dim0dim1 = 'Dim%d-Dim%d' % (first_stack_dimension,
                                    first_stack_dimension + 1)

        self.qcbAxisSelection.addItem(icons.getQIcon("cube-front"), dim1dim2)
        self.qcbAxisSelection.addItem(icons.getQIcon("cube-bottom"), dim0dim2)
        self.qcbAxisSelection.addItem(icons.getQIcon("cube-left"), dim0dim1)

    def setFirstStackDimension(self, first_stack_dim):
        """When viewing the last 3 dimensions of an n-D array (n>3), you can
        use this method to change the text in the combobox.

        For instance, for a 7-D array, first stack dim is 4, so the default
        "Dim1-Dim2" text should be replaced with "Dim5-Dim6" (dimensions
        numbers are 0-based).

        :param int first_stack_dim: First stack dimension (n-3) when viewing the
            last 3 dimensions of an n-D array.
        """
        self._setCBChoices(first_stack_dim)

    def setPerspective(self, perspective):
        """Update the combobox selection.

          - slice plane Dim1-Dim2: perspective 0
          - slice plane Dim0-Dim2: perspective 1
          - slice plane Dim0-Dim1: perspective 2

        :param perspective: Orthogonal dimension number (0, 1, or 2)
        """
        self.qcbAxisSelection.setCurrentIndex(perspective)


class StackViewMainWindow(StackView):
    """This class is a :class:`StackView` with a menu, an additional toolbar
    to set the plot limits, and a status bar to display the value and 3D
    index of the data samples hovered by the mouse cursor.

    :param QWidget parent: Parent widget, or None
    """
    def __init__(self, parent=None):
        self._dataInfo = None
        super(StackViewMainWindow, self).__init__(parent)
        self.setWindowFlags(qt.Qt.Window)

        # Add toolbars and status bar
        self.addToolBar(qt.Qt.BottomToolBarArea,
                        LimitsToolBar(plot=self._plot))

        self.statusBar()

        menu = self.menuBar().addMenu('File')
        menu.addAction(self._plot.getOutputToolBar().getSaveAction())
        menu.addAction(self._plot.getOutputToolBar().getPrintAction())
        menu.addSeparator()
        action = menu.addAction('Quit')
        action.triggered[bool].connect(qt.QApplication.instance().quit)

        menu = self.menuBar().addMenu('Edit')
        menu.addAction(self._plot.getOutputToolBar().getCopyAction())
        menu.addSeparator()
        menu.addAction(self._plot.getResetZoomAction())
        menu.addAction(self._plot.getColormapAction())
        menu.addAction(self.getColorBarAction())

        menu.addAction(actions.control.KeepAspectRatioAction(self._plot, self))
        menu.addAction(actions.control.YAxisInvertedAction(self._plot, self))

        menu = self.menuBar().addMenu('Profile')
        profileToolBar = self._profileToolBar
        menu.addAction(profileToolBar.hLineAction)
        menu.addAction(profileToolBar.vLineAction)
        menu.addAction(profileToolBar.lineAction)
        menu.addAction(profileToolBar.crossAction)
        menu.addSeparator()
        menu.addAction(profileToolBar._editor)
        menu.addSeparator()
        menu.addAction(profileToolBar.clearAction)

        # Connect to StackView's signal
        self.valueChanged.connect(self._statusBarSlot)

    def _statusBarSlot(self, x, y, value):
        """Update status bar with coordinates/value from plots."""
        # todo (after implementing calibration):
        #  - use floats for (x, y, z)
        #  - display both indices (dim0, dim1, dim2) and (x, y, z)
        msg = "Cursor out of range"
        if x is not None and y is not None:
            img_idx = self._browser.value()

            if self._perspective == 0:
                dim0, dim1, dim2 = img_idx, int(y), int(x)
            elif self._perspective == 1:
                dim0, dim1, dim2 = int(y), img_idx, int(x)
            elif self._perspective == 2:
                dim0, dim1, dim2 = int(y), int(x), img_idx

            msg = 'Position: (%d, %d, %d)' % (dim0, dim1, dim2)
        if value is not None:
            msg += ', Value: %g' % value
        if self._dataInfo is not None:
            msg = self._dataInfo + ', ' + msg

        self.statusBar().showMessage(msg)

    def setStack(self, stack, *args, **kwargs):
        """Set the displayed stack.

        See :meth:`StackView.setStack` for details.
        """
        if hasattr(stack, 'dtype') and hasattr(stack, 'shape'):
            assert len(stack.shape) == 3
            nframes, height, width = stack.shape
            self._dataInfo = 'Data: %dx%dx%d (%s)' % (nframes, height, width,
                                                      str(stack.dtype))
            self.statusBar().showMessage(self._dataInfo)
        else:
            self._dataInfo = None

        # Set the new stack in StackView widget
        super(StackViewMainWindow, self).setStack(stack, *args, **kwargs)
        self.setStatusBar(None)



#!/usr/bin/env python
# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Test of the OpenCL array_utils"""

from __future__ import division, print_function

__authors__ = ["Pierre paleo"]
__license__ = "MIT"
__copyright__ = "2013-2017 European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "14/06/2017"


import time
import logging
import numpy as np
import unittest
try:
    import mako
except ImportError:
    mako = None
from ..common import ocl
if ocl:
    import pyopencl as cl
    import pyopencl.array as parray
    from .. import linalg
from ..utils import get_opencl_code
from silx.test.utils import utilstest

logger = logging.getLogger(__name__)
try:
    from scipy.ndimage.filters import laplace
    _has_scipy = True
except ImportError:
    _has_scipy = False



@unittest.skipUnless(ocl and mako, "PyOpenCl is missing")
class TestCpy2d(unittest.TestCase):

    def setUp(self):
        if ocl is None:
            return
        self.ctx = ocl.create_context()
        if logger.getEffectiveLevel() <= logging.INFO:
            self.PROFILE = True
            self.queue = cl.CommandQueue(
                            self.ctx,
                            properties=cl.command_queue_properties.PROFILING_ENABLE)
        else:
            self.PROFILE = False
            self.queue = cl.CommandQueue(self.ctx)
        self.allocate_arrays()
        self.program = cl.Program(self.ctx, get_opencl_code("array_utils")).build()

    def allocate_arrays(self):
        """
        Allocate various types of arrays for the tests
        """
        self.prng_state = np.random.get_state()
        # Generate arrays of random shape
        self.shape1 = np.random.randint(20, high=512, size=(2,))
        self.shape2 = np.random.randint(20, high=512, size=(2,))
        self.array1 = np.random.rand(*self.shape1).astype(np.float32)
        self.array2 = np.random.rand(*self.shape2).astype(np.float32)
        self.d_array1 = parray.to_device(self.queue, self.array1)
        self.d_array2 = parray.to_device(self.queue, self.array2)
        # Generate random offsets
        offset1_y = np.random.randint(2, high=min(self.shape1[0], self.shape2[0]) - 10)
        offset1_x = np.random.randint(2, high=min(self.shape1[1], self.shape2[1]) - 10)
        offset2_y = np.random.randint(2, high=min(self.shape1[0], self.shape2[0]) - 10)
        offset2_x = np.random.randint(2, high=min(self.shape1[1], self.shape2[1]) - 10)
        self.offset1 = (offset1_y, offset1_x)
        self.offset2 = (offset2_y, offset2_x)
        # Compute the size of the rectangle to transfer
        size_y = np.random.randint(2, high=min(self.shape1[0], self.shape2[0]) - max(offset1_y, offset2_y) + 1)
        size_x = np.random.randint(2, high=min(self.shape1[1], self.shape2[1]) - max(offset1_x, offset2_x) + 1)
        self.transfer_shape = (size_y, size_x)

    def tearDown(self):
        self.array1 = None
        self.array2 = None
        self.d_array1.data.release()
        self.d_array2.data.release()
        self.d_array1 = None
        self.d_array2 = None
        self.ctx = None
        self.queue = None

    def compare(self, result, reference):
        errmax = np.max(np.abs(result - reference))
        logger.info("Max error = %e" % (errmax))
        self.assertTrue(errmax == 0, str("Max error is too high"))#. PRNG state was %s" % str(self.prng_state)))

    @unittest.skipUnless(ocl and mako, "pyopencl is missing")
    def test_cpy2d(self):
        """
        Test rectangular transfer of self.d_array1 to self.d_array2
        """
        # Reference
        o1 = self.offset1
        o2 = self.offset2
        T = self.transfer_shape
        logger.info("""Testing D->D rectangular copy with (N1_y, N1_x) = %s,
                    (N2_y, N2_x) = %s:
                    array2[%d:%d, %d:%d] = array1[%d:%d, %d:%d]""" %
                        (
                            str(self.shape1), str(self.shape2),
                            o2[0], o2[0] + T[0],
                            o2[1], o2[1] + T[1],
                            o1[0], o1[0] + T[0],
                            o1[1], o1[1] + T[1]
                        )
                    )
        self.array2[o2[0]:o2[0] + T[0], o2[1]:o2[1] + T[1]] = self.array1[o1[0]:o1[0] + T[0], o1[1]:o1[1] + T[1]]
        kernel_args = (
            self.d_array2.data,
            self.d_array1.data,
            np.int32(self.shape2[1]),
            np.int32(self.shape1[1]),
            np.int32(self.offset2[::-1]),
            np.int32(self.offset1[::-1]),
            np.int32(self.transfer_shape[::-1])
        )
        wg = None
        ndrange = self.transfer_shape[::-1]
        self.program.cpy2d(self.queue, ndrange, wg, *kernel_args)
        res = self.d_array2.get()
        self.compare(res, self.array2)


def suite():
    testSuite = unittest.TestSuite()
    testSuite.addTest(TestCpy2d("test_cpy2d"))
    return testSuite

if __name__ == '__main__':
    unittest.main(defaultTest="suite")
